from uuid import uuid4
from itertools import cycle

import plotly.graph_objs as go
from plotly.colors import DEFAULT_PLOTLY_COLORS

from dash import html, dcc
from dash.dependencies import Output, Input, State
from everviz.plugins.plugin_abc import EvervizPluginABC
from everviz.data.load_csv.get_data import get_data
from everviz.util import parse_range, get_placeholder_text, base64encode

from .util import calculate_statistics


class ObjectivesPlot(EvervizPluginABC):
    """
    The ObjectivesPlot class implements a plugin for Webviz, for plotting the
    objective values generated during an Everest optimization.

    Generally there are multiple realizations for the objective function values
    at each batch. Additionally, there may be multiple objective functions that
    are optimized simultaneously. These can be plotted either all individually
    in the plot, or they are summarized by plotting the mean value together with
    a P10-P90 range.
    """

    def __init__(self, app, csv_file):
        super().__init__()
        self.graph_id = f"graph-{uuid4()}"
        self.radio_id = f"dropdown-{uuid4()}"
        self.radio_id_mode = f"dropdown-{uuid4()}"
        self.function_dropdown_id = f"dropdown-{uuid4()}"
        self.realization_filter_check_id = f"check-{uuid4()}"
        self.realization_filter_input_id = f"input-{uuid4()}"
        self.label_id = f"label-{uuid4()}"
        self.csv_file = csv_file
        self.set_callbacks(app)

    def add_webvizstore(self):
        return [
            (get_data, [{"csv_file": self.csv_file}]),
        ]

    @property
    def layout(self):
        data = get_data(self.csv_file)

        functions = data["function"].unique()
        function_dropdown_options = [{"label": i, "value": i} for i in functions]
        realizations = data["realization"].unique()
        placeholder_text = get_placeholder_text(realizations)
        func_elements = [
            html.Label("Functions to plot:"),
            dcc.Dropdown(
                id=self.function_dropdown_id,
                options=function_dropdown_options,
                multi=True,
                value=[functions[0]],
            ),
        ]

        radio_options = ["Statistics", "Values", "Normalized", "Weighted + Normalized"]
        radio_elements = [
            html.Label(
                f"Statistics are calculated based on {len(realizations)} realizations",
                id=self.label_id,
                style={"display": "none"},
            ),
            dcc.RadioItems(
                id=self.radio_id,
                options=[{"label": i, "value": i} for i in radio_options],
                value=radio_options[1],
                labelStyle={"display": "inline-block"},
                style={"margin-top": 20},
            ),
            html.Label("Style:"),
            dcc.RadioItems(
                id=self.radio_id_mode,
                options=[
                    {"label": "line", "value": "lines"},
                    {"label": "scatter", "value": "markers"},
                ],
                value="lines",
                labelStyle={"display": "inline-block"},
            ),
        ]

        realization_elements = [
            html.Div(
                [
                    dcc.Checklist(
                        id=self.realization_filter_check_id,
                        options=[{"label": "Filter realizations:", "value": "filter"}],
                        style={"display": "inline-block", "margin-right": 8},
                    ),
                    dcc.Input(
                        id=self.realization_filter_input_id,
                        type="text",
                        placeholder=placeholder_text,
                        pattern=r"\s*|([0-9]+(\s*-\s*[0-9]+)?)(\s*,\s*[0-9]+(\s*-\s*[0-9]+)?)*",
                        style={"display": "inline-block"},
                    ),
                ],
                style={"margin-top": 20},
            )
        ]

        return html.Div(
            [
                html.Div(
                    [
                        html.Div(
                            func_elements + radio_elements + realization_elements,
                            style={
                                "width": "29%",
                                "display": "inline-block",
                                "vertical-align": "top",
                            },
                        ),
                        html.Div(
                            [
                                dcc.Graph(
                                    id=self.graph_id,
                                    config={
                                        "displaylogo": False,
                                        "toImageButtonOptions": {
                                            "filename": "objectives"
                                        },
                                    },
                                )
                            ],
                            style={"width": "69%", "display": "inline-block"},
                        ),
                    ]
                ),
            ]
        )

    def set_callbacks(self, app):
        @app.callback(
            self.plugin_data_output,
            [self.plugin_data_requested],
            [
                State(self.radio_id, "value"),
                State(self.realization_filter_check_id, "value"),
                State(self.realization_filter_input_id, "value"),
            ],
        )
        def user_download_data(
            data_requested,
            radio_value,
            realizations_check,
            realizations_input,
        ):
            if data_requested:
                content = get_data(self.csv_file)
                if realizations_check:
                    realizations = parse_range(realizations_input)
                    if realizations:
                        content = content[content["realization"].isin(realizations)]
                if radio_value == "Statistics":
                    filename = "objective_statistics.csv"
                    content = calculate_statistics(content)
                else:
                    filename = "objective_values.csv"
                return {
                    "filename": filename,
                    "content": base64encode(content.to_csv()),
                    "mime_type": "text/csv",
                }
            return None

        @app.callback(
            Output(self.realization_filter_input_id, "disabled"),
            [Input(self.realization_filter_check_id, "value")],
        )
        def set_button_enabled_state(filter_realizations):
            return not filter_realizations

        @app.callback(
            Output(self.label_id, "style"),
            [Input(self.radio_id, "value")],
        )
        def radio_group_click(option):
            if option == "Statistics":
                return {}
            return {"display": "none"}

        @app.callback(
            Output(self.graph_id, "figure"),
            [
                Input(self.function_dropdown_id, "value"),
                Input(self.radio_id, "value"),
                Input(self.radio_id_mode, "value"),
                Input(self.realization_filter_check_id, "value"),
                Input(self.realization_filter_input_id, "value"),
                Input(self.graph_id, "clickData"),
            ],
        )
        def update_graph(
            func_list,
            radio_value,
            plot_mode,
            realizations_check,
            realizations_input,
            click_data,
        ):
            # The key_list arguments is the list of functions to plot.
            if func_list is None or len(func_list) == 0:
                return {}

            # Get the data, filter if requested.
            data = get_data(self.csv_file)
            if realizations_check:
                realizations = parse_range(realizations_input)
                if realizations:
                    data = data[data["realization"].isin(realizations)]
            if len(data) == 0:
                return {}

            if radio_value == "Statistics":
                data = calculate_statistics(data)

            data = data.set_index("function")

            # Make a cycle iterator over the plotly colors.
            colors = cycle(DEFAULT_PLOTLY_COLORS)

            traces = []
            for color, key in zip(colors, func_list):
                # Select all data belonging to the current key.
                func_data = data.xs(key)

                # Make the traces, with mean, P10 and P90, shading in between.
                if radio_value == "Statistics":
                    traces.extend(statistics_traces(color, func_data, key, plot_mode))
                else:
                    traces.extend(
                        function_traces(
                            func_list,
                            click_data,
                            color,
                            func_data,
                            key,
                            plot_mode,
                            radio_value,
                        )
                    )

            return {
                "data": traces,
                "layout": dict(
                    xaxis={"title": "Batch", "tickformat": ",d"},
                    yaxis={"title": "Function Key Value"},
                    hovermode="closest",
                ),
            }


def statistics_traces(color, func_data, key, plot_mode):
    return [
        go.Scattergl(
            y=func_data["P90"],
            x=func_data["batch"],
            mode=plot_mode,
            marker={"size": 10},
            line={"color": color},
            name=key + "(P90)",
            showlegend=False,
        ),
        go.Scattergl(
            y=func_data["P10"],
            x=func_data["batch"],
            mode=plot_mode,
            line={"color": color},
            marker={"size": 10},
            fill="tonexty",
            name=key + "(P10)",
            showlegend=False,
        ),
        go.Scattergl(
            y=func_data["mean"],
            x=func_data["batch"],
            mode=plot_mode,
            marker={"size": 10},
            line={"color": color},
            name=key,
            showlegend=True,
        ),
        go.Scattergl(
            y=func_data["min_value"],
            x=func_data["batch"],
            mode="markers",
            marker={"size": 10, "symbol": "square"},
            line={"color": color},
            name=key,
            showlegend=False,
            hovertext=[f"realization: {r}" for r in func_data["min_realization"]],
        ),
        go.Scattergl(
            y=func_data["max_value"],
            x=func_data["batch"],
            mode="markers",
            marker={"size": 10, "symbol": "square"},
            line={"color": color},
            name=key,
            showlegend=False,
            hovertext=[f"realization: {r}" for r in func_data["max_realization"]],
        ),
    ]


def function_traces(
    func_list, click_data, color, func_data, key, plot_mode, radio_value
):
    realizations = func_data["realization"].unique()
    for real in realizations:
        line = func_data[func_data["realization"].isin([real])]
        y_values = line["value"]
        if radio_value == "Normalized":
            y_values = y_values * line["norm"]
        elif radio_value == "Weighted + Normalized":
            y_values = y_values * line["norm"] * line["weight"]
        line_style = {"dash": "dash"}
        if len(func_list) > 1:
            line_style["color"] = color

        if click_data is not None:
            clicked_realization = click_data["points"][0].get("customdata", None)
            if clicked_realization == real:
                line_style["width"] = 4
                del line_style["dash"]

        mode = plot_mode
        if mode == "lines":
            mode = "lines+markers"

        yield go.Scattergl(
            y=y_values,
            x=line["batch"],
            mode=mode,
            marker={"size": 10},
            name=f"{key}, r={real}",
            customdata=[real] * len(line["value"]),
            showlegend=True,
            line=line_style,
        )



import os

import pandas as pd
import numpy as np

import everviz

from everviz.pages.objectives import (
    _set_up_data_sources,
    _objective_values,
    _total_objective_values,
)

from everviz.plugins.objectives_plot.util import calculate_statistics

OBJECTIVES = [
    {
        "batch": 0,
        "realization": 1,
        "function": "f0",
        "value": 100,
        "simulation": 1,
        "weight": 0.75,
        "norm": 1,
    },
    {
        "batch": 0,
        "realization": 1,
        "function": "f1",
        "value": 200,
        "simulation": 1,
        "weight": 0.25,
        "norm": 1.5,
    },
    {
        "batch": 2,
        "realization": 2,
        "function": "f0",
        "value": 200,
        "simulation": 2,
        "weight": 0.75,
        "norm": 1,
    },
    {
        "batch": 2,
        "realization": 2,
        "function": "f1",
        "value": 400,
        "simulation": 2,
        "weight": 0.25,
        "norm": 1.5,
    },
    {
        "batch": 0,
        "realization": 1,
        "function": "f0",
        "value": 300,
        "simulation": 1,
        "weight": 0.75,
        "norm": 1,
    },
    {
        "batch": 0,
        "realization": 1,
        "function": "f1",
        "value": 600,
        "simulation": 1,
        "weight": 0.25,
        "norm": 1.5,
    },
    {
        "batch": 2,
        "realization": 2,
        "function": "f0",
        "value": 400,
        "simulation": 2,
        "weight": 0.75,
        "norm": 1,
    },
    {
        "batch": 2,
        "realization": 2,
        "function": "f1",
        "value": 800,
        "simulation": 2,
        "weight": 0.25,
        "norm": 1.5,
    },
]


def test_objective_values_data_frame():
    """Test for the correct layout and size of the objective values data frame."""
    objective_values = _objective_values(pd.DataFrame(OBJECTIVES))

    assert set(objective_values.columns) == {
        "batch",
        "function",
        "value",
        "realization",
        "weight",
        "norm",
    }
    assert len(objective_values) == len(OBJECTIVES)
    assert set(objective_values["batch"]) == {obj["batch"] for obj in OBJECTIVES}
    assert set(objective_values["function"]) == {obj["function"] for obj in OBJECTIVES}
    assert set(objective_values["realization"]) == {
        obj["realization"] for obj in OBJECTIVES
    }
    assert set(objective_values["weight"]) == {obj["weight"] for obj in OBJECTIVES}
    assert set(objective_values["norm"]) == {obj["norm"] for obj in OBJECTIVES}


def test_objective_statistics_data_frame():
    """Test for the correct layout and size of the objective statistics data frame"""
    objective_values = _objective_values(pd.DataFrame(OBJECTIVES))
    objective_statistics = calculate_statistics(objective_values)

    assert set(objective_statistics.columns) == {
        "batch",
        "function",
        "mean",
        "P10",
        "P90",
        "min_value",
        "max_value",
        "min_realization",
        "max_realization",
    }
    assert len(objective_statistics) == 2 * 2  # #batches * #functions
    assert set(objective_statistics["batch"]) == {obj["batch"] for obj in OBJECTIVES}
    assert set(objective_statistics["function"]) == {
        obj["function"] for obj in OBJECTIVES
    }


def test_objective_statistics_contenIt():
    """Test for the correct content of the objectives statistics data frame."""
    objective_values = _objective_values(pd.DataFrame(OBJECTIVES))
    objective_statistics = calculate_statistics(objective_values)
    mean = []
    p10 = []
    p90 = []
    min_value = []
    max_value = []
    min_realization = []
    max_realization = []
    for func in ["f0", "f1"]:
        for batch in [0, 2]:
            values = [
                obj["value"]
                for obj in OBJECTIVES
                if obj["batch"] == batch and obj["function"] == func
            ]
            realizations = [
                obj["realization"]
                for obj in OBJECTIVES
                if obj["batch"] == batch and obj["function"] == func
            ]
            mean.append(np.mean(values))
            p10.append(np.quantile(values, q=0.1))
            p90.append(np.quantile(values, q=0.9))
            min_value.append(np.min(values))
            max_value.append(np.max(values))
            min_realization.append(realizations[np.argmin(values)])
            max_realization.append(realizations[np.argmax(values)])
    assert objective_statistics["mean"].to_list() == mean
    assert objective_statistics["P10"].to_list() == p10
    assert objective_statistics["P90"].to_list() == p90
    assert objective_statistics["min_value"].to_list() == min_value
    assert objective_statistics["max_value"].to_list() == max_value
    assert objective_statistics["min_realization"].to_list() == min_realization
    assert objective_statistics["max_realization"].to_list() == max_realization


def test_set_up_sources(mocker, monkeypatch, tmpdir):
    mock_api = mocker.Mock()
    mock_api.output_folder = tmpdir
    os.mkdir(os.path.join(tmpdir, "everviz"))
    monkeypatch.setattr(
        everviz.pages.objectives,
        "_objective_values_from_api",
        mocker.Mock(return_value=pd.DataFrame(OBJECTIVES)),
    )
    total_data = pd.DataFrame(OBJECTIVES).drop(
        columns=["realization", "function", "simulation"]
    )
    monkeypatch.setattr(
        everviz.pages.objectives,
        "_total_objective_values_from_api",
        mocker.Mock(return_value=total_data),
    )
    data_source = _set_up_data_sources(mock_api)
    assert data_source.objective_values == tmpdir / "everviz" / "objective_values.csv"
    assert (
        data_source.total_objective_values
        == tmpdir / "everviz" / "total_objective_values.csv"
    )


def test_total_objective_values_data_frame():
    total_data = pd.DataFrame(OBJECTIVES).drop(
        columns=["realization", "function", "simulation", "weight", "norm"]
    )
    total_objective_values = _total_objective_values(total_data)

    assert set(total_objective_values.columns) == {"batch", "value"}
    assert len(total_objective_values) == len(OBJECTIVES)



import os
from everviz.util import get_everviz_folder, DEFAULT_CONFIG


def page_layout(api):
    everviz_path = get_everviz_folder(api)
    config_path = os.path.join(everviz_path, DEFAULT_CONFIG)
    return {
        "title": "Config editor",
        "content": [
            {
                "ConfigEditor": {
                    "data_path": config_path,
                },
            },
        ],
    }



import time
import os
import pytest

import pandas as pd
from PIL import Image

import everviz
from everviz.plugins.objectives_plot.single_objectives_plot import SingleObjectivesPlot


def test_single_objective_plot_layout(
    app, dash_duo, mocker, caplog, tmpdir, assert_equal_images
):
    test_data = [
        {"batch": 0, "objective": 100, "accepted": 1, "npv": 110, "rf": 130},
        {"batch": 1, "objective": 200, "accepted": 1, "npv": 210, "rf": 230},
        {"batch": 2, "objective": 200, "accepted": 0, "npv": 210, "rf": 230},
        {"batch": 3, "objective": 400, "accepted": 1, "npv": 410, "rf": 430},
        {"batch": 4, "objective": 300, "accepted": 0, "npv": 310, "rf": 330},
        {"batch": 5, "objective": 600, "accepted": 1, "npv": 610, "rf": 630},
        {"batch": 6, "objective": 400, "accepted": 0, "npv": 410, "rf": 430},
        {"batch": 7, "objective": 800, "accepted": 1, "npv": 810, "rf": 830},
    ]

    def mock_get_data(_):
        return pd.DataFrame(test_data)

    mocker.patch(
        "everviz.plugins.objectives_plot.single_objectives_plot.get_data",
        side_effect=mock_get_data,
    )

    plugin = SingleObjectivesPlot(app, "csv_data_file")
    app.layout = plugin.layout
    dash_duo.driver.set_window_size(1024, 768)
    dash_duo.start_server(app)

    dash_duo.select_dcc_dropdown("#{}".format(plugin.function_dropdown_id), "npv")
    dash_duo.select_dcc_dropdown("#{}".format(plugin.function_dropdown_id), "rf")

    ref_image_path = os.path.join(
        everviz.__path__[0],
        "..",
        "test-data",
        "objectives",
        "single_objective_layout_headless.png",
    )
    ref_image = Image.open(ref_image_path)

    with tmpdir.as_cwd():
        time.sleep(10)
        dash_duo.driver.save_screenshot("example_snapshot.png")
        screen_shot = Image.open("example_snapshot.png")

    if screen_shot.size != ref_image.size:
        pytest.skip("Reference image size does not match {}".format(screen_shot.size))
    assert_equal_images(ref_image, screen_shot, threshold=1.0)

    dash_duo.clear_input("#{}".format(plugin.function_dropdown_id))

    for record in caplog.records:
        assert record.levelname != "ERROR"



from datetime import datetime

import pandas as pd
import numpy as np

from everviz.pages.summary_values import page_layout, _summary_values
from everviz.plugins.summary_plot.util import calculate_statistics

__TEST_DATA = {
    "realization": range(12),
    "simulation": range(12),
    "batch": [0] * 6 + [1] * 6,
    "date": [
        datetime(2000, 1, 1),
        datetime(2000, 2, 1),
        datetime(2000, 3, 1),
        datetime(2000, 1, 1),
        datetime(2000, 2, 1),
        datetime(2000, 3, 1),
        datetime(2000, 1, 1),
        datetime(2000, 2, 1),
        datetime(2000, 3, 1),
        datetime(2000, 1, 1),
        datetime(2000, 2, 1),
        datetime(2000, 3, 1),
    ],
    "key1": range(1, 13),
    "key2": range(10, 130, 10),
}


def test_empty_summary_values(mocker):
    """Test for the case of missing summary values."""
    mock_api = mocker.Mock()
    mock_api.output_folder = "dummy"
    mock_api.summary_values.return_value = pd.DataFrame()
    layout = page_layout(mock_api)
    assert not layout


def test_summary_values_data_frame():
    """Test for the correct layout and size of the summary values data frame."""
    summary_values = _summary_values(pd.DataFrame(__TEST_DATA))

    assert list(
        summary_values.columns == ["summary_key", "batch", "date", "key1", "key2"]
    )
    assert len(summary_values) == len(__TEST_DATA["simulation"])
    assert set(summary_values["batch"]) == set(__TEST_DATA["batch"])
    assert set(summary_values["date"]) == {
        pd.Timestamp(date) for date in __TEST_DATA["date"]
    }


def test_summary_statistics_data_frame():
    """Test for the correct layout and size of the summary statistics data frame"""
    summary_values = _summary_values(pd.DataFrame(__TEST_DATA))
    summary_statistics = calculate_statistics(summary_values, ["key1", "key2"])

    assert list(summary_statistics.columns) == [
        "summary_key",
        "batch",
        "date",
        "mean",
        "P10",
        "P90",
        "min_value",
        "max_value",
        "min_realization",
        "max_realization",
    ]
    assert len(summary_statistics) == 2 * 3 * 2  # #batches * #dates * #keys
    assert set(summary_statistics["summary_key"]) == set(["key1", "key2"])
    assert set(summary_statistics["batch"]) == set(__TEST_DATA["batch"])
    assert set(summary_statistics["date"]) == {
        pd.Timestamp(date) for date in __TEST_DATA["date"]
    }


def test_summary_statistics_content():
    """Test for the correct content of the summary statistics data frame."""
    summary_values = _summary_values(pd.DataFrame(__TEST_DATA))
    summary_statistics = calculate_statistics(summary_values, ["key1", "key2"])

    mean = []
    p10 = []
    p90 = []
    min_value = []
    max_value = []
    min_realization = []
    max_realization = []
    for key in ["key1", "key2"]:
        zipped = list(
            zip(
                __TEST_DATA["batch"],
                __TEST_DATA["date"],
                __TEST_DATA[key],
                __TEST_DATA["realization"],
            )
        )
        for batch in [0, 1]:
            for date in [
                datetime(2000, 1, 1),
                datetime(2000, 2, 1),
                datetime(2000, 3, 1),
            ]:
                values = [v for b, d, v, _ in zipped if batch == b and date == d]
                realizations = [r for b, d, _, r in zipped if batch == b and date == d]
                mean.append(np.mean(values))
                p10.append(np.quantile(values, q=0.1))
                p90.append(np.quantile(values, q=0.9))
                min_value.append(np.min(values))
                max_value.append(np.max(values))
                min_realization.append(realizations[np.argmin(values)])
                max_realization.append(realizations[np.argmax(values)])

    assert summary_statistics["mean"].to_list() == mean
    assert summary_statistics["P10"].to_list() == p10
    assert summary_statistics["P90"].to_list() == p90
    assert summary_statistics["min_value"].to_list() == min_value
    assert summary_statistics["max_value"].to_list() == max_value
    assert summary_statistics["min_realization"].to_list() == min_realization
    assert summary_statistics["max_realization"].to_list() == max_realization



from datetime import datetime

import pandas as pd
import pytest

from everviz.plugins.summary_plot.wells_plot import WellsPlot, _get_target_keys
from everviz.pages.summary_values import _summary_values


def test_well_plot(app, dash_duo, mocker, caplog):
    test_data = {
        "realization": range(12),
        "simulation": range(12),
        "batch": [0] * 6 + [1] * 6,
        "date": [
            datetime(2000, 1, 1),
            datetime(2000, 2, 1),
            datetime(2000, 3, 1),
            datetime(2000, 1, 1),
            datetime(2000, 2, 1),
            datetime(2000, 3, 1),
            datetime(2000, 1, 1),
            datetime(2000, 2, 1),
            datetime(2000, 3, 1),
            datetime(2000, 1, 1),
            datetime(2000, 2, 1),
            datetime(2000, 3, 1),
        ],
        "WOPR:OP1": range(1, 13),
        "WOPRT:OP1": range(10, 130, 10),
    }

    mocker.patch(
        "everviz.plugins.summary_plot.wells_plot.get_data",
        return_value=_summary_values(pd.DataFrame(test_data)),
    )

    plugin = WellsPlot(app, "values")
    app.layout = plugin.layout
    dash_duo.start_server(app)

    # Tests warning label not there
    dash_duo.wait_for_contains_text(f"#{plugin.label_id}", "")

    dash_duo.clear_input("#{}".format(plugin.dropdown_key))
    dash_duo.select_dcc_dropdown("#{}".format(plugin.dropdown_key), "WOPR:OP1")
    dash_duo.find_element(
        "#{} label:nth-child({})".format(plugin.radio_statistics, 2)
    ).click()

    # Tests warning label is there
    dash_duo.wait_for_contains_text(f"#{plugin.label_id}", "Statistics are calculated")

    for record in caplog.records:
        assert record.levelname != "ERROR"


@pytest.mark.parametrize(
    "summary_keys, expected",
    (
        (["WOPR:OP1"], ["WOPRT:OP1"]),
        (["WOPR:OP2"], []),
        (["WOPR:OP1", "WOPR:OP2", "WWIR:IN1"], ["WOPRT:OP1", "WWIRT:IN1"]),
    ),
)
def test_get_target_keys(summary_keys, expected):
    df = pd.DataFrame(columns=["WOPRT:OP1", "WWIRT:IN1"])
    assert _get_target_keys(df, summary_keys) == expected



from pathlib import Path
import abc
import pkg_resources
from webviz_config import WebvizPluginABC
from webviz_config.webviz_assets import WEBVIZ_ASSETS


class EvervizPluginABC(WebvizPluginABC):
    """Subclass of the WebvizPluginABC base class
    All everviz-webviz plugins should be subclasses of this base class
    e.g.
    ```python
    class MyPlugin(EvervizPluginABC):

        def __init__(self):
            super().__init__()

    ```
    """

    def __init__(self):
        super().__init__()
        self._screenshot_filename = f"{self.plugin_name()}.png"
        ASSETS_DIR = Path(pkg_resources.resource_filename("everviz", "assets"))
        WEBVIZ_ASSETS.add(ASSETS_DIR / "axis_customization.css")

    def plugin_name(self):
        name = map(lambda c: f"_{c.lower()}" if c.isupper() else c, type(self).__name__)
        return "everviz" + "".join(name)

    @property
    @abc.abstractmethod
    def layout(self):
        """This is the only required function of a Webviz plugin.
        It returns a Dash layout which by webviz-config is added to
        the main Webviz application.
        """



import os
import time

import dash
import pytest
from dash import html
import pandas as pd
from PIL import Image

import everviz
from everviz.plugins import Crossplot
from everviz.pages import crossplot
from everviz.plugins.utils.layout.sidebar_layout import (
    get_sidebar_layout,
    _get_dropdown,
    _get_radio,
)


@pytest.mark.parametrize("multi", [True, False])
@pytest.mark.parametrize(
    "test_input,expected",
    [
        (["dropdown_val"], "dropdown_val"),
        (["1", "2"], "1"),
        (["1", "2"], "2"),
        (None, "Select"),
    ],
)
def test_crossplot_layout_dropdown(dash_duo, test_input, expected, multi):
    app = dash.Dash(__name__)
    item_id = "id"
    layout = _get_dropdown(item_id=item_id, options=test_input, multi=multi)
    app.layout = html.Div(layout)
    dash_duo.start_server(app)

    dash_duo.clear_input("#{}".format(item_id))
    dash_duo.select_dcc_dropdown("#{}".format(item_id), expected)
    result = dash_duo.find_element("#{}".format(item_id)).text.split()[0]

    # If multi is set the text will have an × in front in the dropdown text
    if result.startswith("×"):
        result = result[1:]

    assert result == expected


@pytest.mark.parametrize(
    "valid_options", [[("OPT_1", "nr_1"), ("OPT_2", "nr_2")], ["OPT_1", "OPT_2"]]
)
@pytest.mark.parametrize("expected, placement", [("OPT_1", 1), ("OPT_2", 2)])
def test_crossplot_layout_radio(dash_duo, expected, placement, valid_options):
    app = dash.Dash(__name__)
    layout = _get_radio("radio_id", options=valid_options)
    app.layout = html.Div(layout)
    dash_duo.start_server(app)

    text = dash_duo.find_element(
        "#{} label:nth-child({})".format("radio_id", placement)
    ).text
    assert text == expected


def test_get_sidebar_layout(monkeypatch, mocker):
    radio_mock = mocker.Mock(return_value=[])
    dropdown_mock = mocker.Mock(return_value=[])
    label_mock = mocker.Mock(return_value=[])
    monkeypatch.setattr(
        everviz.plugins.utils.layout.sidebar_layout, "_get_radio", radio_mock
    )
    monkeypatch.setattr(
        everviz.plugins.utils.layout.sidebar_layout,
        "_get_dropdown",
        dropdown_mock,
    )
    monkeypatch.setattr(
        everviz.plugins.utils.layout.sidebar_layout,
        "_get_label",
        label_mock,
    )

    get_sidebar_layout(
        [
            ("label", {"item_id": "label_id"}),
            ("radio", {"item_id": "radio_id"}),
            ("dropdown", {"item_id": "dropdown_id"}),
        ]
    )

    radio_mock.assert_called_once_with(item_id="radio_id")
    dropdown_mock.assert_called_once_with(item_id="dropdown_id")
    label_mock.assert_called_once_with(item_id="label_id")


def test_crossplot_layout(dash_duo, monkeypatch, mocker, tmpdir, assert_equal_images):
    app = dash.Dash(__name__)
    mock_data = pd.DataFrame(data=[[1, 2, 3]], columns=["a", "b", "c"])
    monkeypatch.setattr(
        everviz.plugins.crossplot.crossplot,
        "get_data",
        mocker.Mock(return_value=mock_data),
    )
    monkeypatch.setattr(
        everviz.plugins.crossplot.crossplot.Crossplot, "set_callbacks", mocker.Mock()
    )

    plugin = Crossplot(mocker.Mock(), "data_path")
    layout = plugin.layout
    app.layout = layout
    dash_duo.start_server(app)

    _REFERENCE_IMAGES = {
        (1024, 768): "crossplot_layout_headless.png",
    }

    dash_duo.driver.set_window_size(1024, 768)
    with tmpdir.as_cwd():
        # Graph can be slow to load, so we need to ensure it has loaded.
        time.sleep(10)
        dash_duo.driver.save_screenshot("example_snapshot.png")
        snapshot = Image.open("example_snapshot.png")

    if snapshot.size not in _REFERENCE_IMAGES:
        pytest.skip("No reference image for layout size: {}".format(snapshot.size))

    reference_image = Image.open(
        os.path.join(
            everviz.__path__[0],
            "..",
            "test-data",
            "crossplot",
            _REFERENCE_IMAGES[snapshot.size],
        )
    )

    assert_equal_images(reference_image, snapshot, threshold=1.0)


def test_webviz_page_layout(mocker):
    api_mock = mocker.Mock()
    api_mock.everest_csv = "everest_export.csv"
    api_mock.control_names = ["x-0", "x-1"]
    result = crossplot.page_layout(api_mock)
    expected_page_layout = {
        "title": "Cross plots",
        "content": [
            crossplot._crossplot_doc,
            {
                "Crossplot": {
                    "data_path": "everest_export.csv",
                },
            },
            crossplot._crossplot_indexed_doc,
            {
                "CrossplotIndexed": {
                    "data_path": "everest_export.csv",
                }
            },
        ],
    }
    assert expected_page_layout == result


def test_webviz_page_layout_no_crossplot(mocker):
    api_mock = mocker.Mock()
    api_mock.everest_csv = "everest_export.csv"
    api_mock.control_names = ["x", "y"]
    result = crossplot.page_layout(api_mock)
    expected_page_layout = {
        "title": "Cross plots",
        "content": [
            crossplot._crossplot_doc,
            {
                "Crossplot": {
                    "data_path": "everest_export.csv",
                },
            },
        ],
    }
    assert expected_page_layout == result


def test_webviz_page_layout_nothing(mocker):
    api_mock = mocker.Mock()
    api_mock.control_names = ["x-0", "x-1"]
    result = crossplot.page_layout(api_mock)
    api_mock.everest_csv = None
    result = crossplot.page_layout(api_mock)
    expected_page_layout = ""
    assert expected_page_layout == result



from everviz.plugins.crossplot.crossplot import Crossplot
from everviz.plugins.crossplot.crossplot_indexed import CrossplotIndexed
from everviz.plugins.summary_plot.summary_plot import SummaryPlot
from everviz.plugins.objectives_plot.objectives_plot import ObjectivesPlot
from everviz.plugins.objectives_plot.single_objectives_plot import SingleObjectivesPlot
from everviz.plugins.controls_plot.controls_plot import ControlsPlot
from everviz.plugins.best_controls_plot.best_controls_plot import BestControlsPlot
from everviz.plugins.delta_plot.delta_plot import DeltaPlot
from everviz.plugins.gradient_plot import GradientPlot
from everviz.plugins.config_editor.config_editor import ConfigEditor
from everviz.plugins.summary_plot.wells_plot import WellsPlot

__all__ = [
    "Crossplot",
    "CrossplotIndexed",
    "SummaryPlot",
    "ObjectivesPlot",
    "SingleObjectivesPlot",
    "ControlsPlot",
    "BestControlsPlot",
    "DeltaPlot",
    "GradientPlot",
    "ConfigEditor",
    "WellsPlot",
]



import pytest

from everviz.util import identify_indexed_controls


@pytest.mark.parametrize(
    "test_input, expected",
    [
        (["base_name-0", "base_name-1"], {"base_name": [0, 1]}),
        (["base_name-0", "indexed-0"], {"base_name": [0], "indexed": [0]}),
        (["not_indexed"], {}),
        (["not_indexed-a"], {}),
    ],
)
def test_get_indexed_controls(test_input, expected):
    result = identify_indexed_controls(test_input)
    assert result == expected



import pandas as pd
from everviz.plugins.gradient_plot import GradientPlot


_DATA = [
    {"batch": 0, "function": "f1", "control": "c1", "value": 1},
    {"batch": 0, "function": "f1", "control": "c2", "value": 2},
    {"batch": 0, "function": "f2", "control": "c1", "value": 3},
    {"batch": 0, "function": "f2", "control": "c2", "value": 4},
    {"batch": 1, "function": "f1", "control": "c1", "value": 5},
    {"batch": 1, "function": "f1", "control": "c2", "value": 6},
    {"batch": 1, "function": "f2", "control": "c1", "value": 7},
    {"batch": 1, "function": "f2", "control": "c2", "value": 8},
]


def test_gradient_plot_callback(app, dash_duo, mocker, caplog, helpers):
    mocker.patch(
        "everviz.plugins.gradient_plot.get_data",
        return_value=pd.DataFrame(_DATA),
    )

    plugin = GradientPlot(app, "dummy")
    app.layout = plugin.layout
    dash_duo.start_server(app)

    control_list_id = f"#multi-selector_id_{plugin.list_select_id}"
    control_selected_id = f"#selected_dropdown_id_{plugin.list_select_id}"

    dash_duo.wait_for_contains_text(control_list_id, "c1\nc2", timeout=5)

    first_control = helpers.select_first(dash_duo, control_list_id)
    dash_duo.wait_for_text_to_equal(control_list_id, "c2", timeout=5)

    second_control = helpers.select_first(dash_duo, control_list_id)
    assert {first_control, second_control} == {"c1", "c2"}
    dash_duo.wait_for_text_to_equal(control_list_id, "", timeout=5)

    # Test adding a function.
    dash_duo.select_dcc_dropdown("#{}".format(plugin.function_dropdown_id), "f2")

    # Clear the dropdown, which should not cause an error. Call twice to
    # simulate two clear key presses.
    dash_duo.clear_input("#{}".format(plugin.function_dropdown_id))
    dash_duo.clear_input("#{}".format(plugin.function_dropdown_id))

    # Test adding a batch
    dash_duo.select_dcc_dropdown("#{}".format(plugin.function_dropdown_id), "f2")
    dash_duo.select_dcc_dropdown("#{}".format(plugin.batch_dropdown_id), "1")

    # Clear the dropdown, which should not cause an error. Call twice to
    # simulate two clear key presses.
    dash_duo.clear_input("#{}".format(plugin.batch_dropdown_id))
    dash_duo.clear_input("#{}".format(plugin.batch_dropdown_id))

    # Test normalization radio.
    dash_duo.select_dcc_dropdown("#{}".format(plugin.batch_dropdown_id), "0")
    dash_duo.find_element(
        "#{} label:nth-child({})".format(plugin.normalization_radio_id, 2)
    ).click()
    dash_duo.find_element(
        "#{} label:nth-child({})".format(plugin.normalization_radio_id, 3)
    ).click()

    # Test abs check.
    dash_duo.find_element(
        "#{} label:nth-child({})".format(plugin.abs_check_id, 1)
    ).click()

    helpers.clear_dropdown(dash_duo, control_selected_id)
    dash_duo.wait_for_contains_text(control_list_id, "c1\nc2", timeout=5)

    for record in caplog.records:
        assert record.levelname != "ERROR"



import os
import sys
import pluggy
import pytest
import dash
from selenium.webdriver.common.by import By

from PIL import ImageFilter, ImageChops

# Allow everviz to be installed and have tests run without installing everest
module = type(sys)("everest.plugins")
module.hookimpl = pluggy.HookimplMarker("test")
sys.modules["everest.plugins"] = module


@pytest.fixture()
def assert_equal_images(tmpdir):
    def func(image_1, image_2, threshold=0.1):
        """
        In this function we check if two images are almost identical
         by blurring and calculating MSE between the images.
        """
        filename1 = (
            os.path.join(tmpdir, image_1.filename)
            if not os.path.isabs(image_1.filename)
            else image_1.filename
        )
        filename2 = (
            os.path.join(tmpdir, image_2.filename)
            if not os.path.isabs(image_2.filename)
            else image_2.filename
        )

        image_1, image_2 = [image.convert("L") for image in [image_1, image_2]]
        image_1, image_2 = [
            image.filter(ImageFilter.GaussianBlur(radius=5))
            for image in [image_1, image_2]
        ]

        pixel_pairs = zip(list(image_1.getdata()), list(image_2.getdata()))
        diff = sum((pixel_1 - pixel_2) ** 2 for pixel_1, pixel_2 in pixel_pairs)

        error = diff / (image_1.size[0] * image_1.size[1])

        assert (
            error < threshold
        ), "Bounding box: {}\nImage 1 path: {}\nImage 2 path: {}\n".format(
            ImageChops.subtract(image_1, image_2).getbbox(), filename1, filename2
        )

    return func


@pytest.fixture()
def app():
    dash_app = dash.Dash(__name__)
    dash_app.config.suppress_callback_exceptions = True
    yield dash_app


class HelperFunctions:
    """Class containing helper functions used in tests"""

    @staticmethod
    def select_first(dash_duo, selector):
        element = dash_duo.find_element(selector)
        option = element.find_element(By.CSS_SELECTOR, "option")
        option_text = option.text
        option.click()
        return option_text

    @staticmethod
    def clear_dropdown(dash_duo, selector):
        element = dash_duo.find_element(selector)
        element.find_element(By.CSS_SELECTOR, "span.Select-clear").click()


@pytest.fixture
def helpers():
    return HelperFunctions



import os
from collections import namedtuple
import pandas as pd
from everviz.log import get_logger


DataSources = namedtuple("DataSource", "controls_per_batch, controls_initial_vs_best")

logger = get_logger()


def _control_data_per_batch(api):
    return pd.DataFrame(api.control_values)


def _control_data_initial_vs_best(api):
    objectives_df = pd.DataFrame(api.single_objective_values)
    best_batch = objectives_df.batch[objectives_df.objective.idxmax()]
    data = pd.DataFrame(api.control_values)

    # Keep only controls associated with initial and best batches
    data = data[(data.batch == 0) | (data.batch == best_batch)]
    data = data.replace({"batch": 0}, "initial").replace({"batch": best_batch}, "best")
    return data


def _set_up_data_sources(api):
    everest_folder = api.output_folder
    everviz_path = os.path.join(everest_folder, "everviz")

    logger.info("Generating controls per batch source data file")
    controls_per_batch = os.path.join(everviz_path, "controls_per_batch.csv")
    data = _control_data_per_batch(api)
    data.to_csv(controls_per_batch, index=False)
    logger.info(f"File created: {controls_per_batch}")  # pylint: disable=W1203

    logger.info("Generating initial vs best controls source data file")
    controls_initial_vs_best = os.path.join(
        everviz_path, "controls_initial_vs_best.csv"
    )
    data = _control_data_initial_vs_best(api)
    data.to_csv(controls_initial_vs_best, index=False)

    return DataSources(
        controls_per_batch=controls_per_batch,
        controls_initial_vs_best=controls_initial_vs_best,
    )


def page_layout(api):
    sources = _set_up_data_sources(api)
    return {
        "title": "Controls",
        "content": [
            "## Control value per batch",
            {"ControlsPlot": {"csv_file": sources.controls_per_batch}},
            "## Initial controls versus best controls",
            {
                "BestControlsPlot": {
                    "csv_file": sources.controls_initial_vs_best,
                }
            },
        ],
    }



import os
from collections import namedtuple

from everviz.log import get_logger

WELLS_EXPLANATION = """
The well plot requires well summary data to be loaded from the forward model.
Make sure the wells are specified in the Everest config file and exported from
Eclipse/Flow to *.UNSMRY. To visualize target rates they also need to be added 
to the simulator summary data section.
"""

DataSources = namedtuple("DataSource", ["summary_values"])

logger = get_logger()


def _set_up_data_sources(api):
    everest_folder = api.output_folder
    everviz_path = os.path.join(everest_folder, "everviz")

    summary_values_file = os.path.join(everviz_path, "summary_values.csv")

    if not os.path.exists(summary_values_file):
        return None

    return DataSources(summary_values=summary_values_file)


def page_layout(api):
    sources = _set_up_data_sources(api)

    if sources is None:
        return {
            "title": "Well rates",
            "content": [
                WELLS_EXPLANATION,
            ],
        }

    return {
        "title": "Well rates",
        "content": [
            "## Well rate values as a function of date",
            {
                "WellsPlot": {
                    "csv_file": sources.summary_values,
                },
            },
        ],
    }



import os

import pytest
import numpy as np
import pandas as pd

from everviz.util import get_everviz_folder
from everviz.pages.controls import (
    _control_data_per_batch,
    _control_data_initial_vs_best,
    DataSources,
    _set_up_data_sources,
)


controls = [
    [
        {"control": "c1", "batch": 0, "value": 12},
    ],
    [
        {"control": "c1", "batch": 0, "value": 12},
        {"control": "c1", "batch": 1, "value": 30},
    ],
    [
        {"control": "c1", "batch": 0, "value": 12},
        {"control": "c2", "batch": 0, "value": 30},
        {"control": "c1", "batch": 1, "value": 20},
        {"control": "c2", "batch": 1, "value": 22},
        {"control": "c1", "batch": 2, "value": 10},  # best
        {"control": "c2", "batch": 2, "value": 32},  # best
        {"control": "c1", "batch": 3, "value": 2},
        {"control": "c2", "batch": 3, "value": 40},
    ],
]
objectives = [
    [
        {
            "batch": 0,
            "objective": 1,
        },
    ],
    [
        {
            "batch": 0,
            "objective": 1,
        },
        {
            "batch": 1,
            "objective": 2,
        },
    ],
    [
        {
            "batch": 0,
            "objective": 1,
        },
        {
            "batch": 1,
            "objective": 2,
        },
        {
            "batch": 2,
            "objective": 4,
        },  # best value
        {
            "batch": 3,
            "objective": 0,
        },
    ],
]

expected = [
    [
        {"control": "c1", "batch": "initial", "value": 12},
    ],
    [
        {"control": "c1", "batch": "initial", "value": 12},
        {"control": "c1", "batch": "best", "value": 30},
    ],
    [
        {"control": "c1", "batch": "initial", "value": 12},
        {"control": "c2", "batch": "initial", "value": 30},
        {"control": "c1", "batch": "best", "value": 10},
        {"control": "c2", "batch": "best", "value": 32},
    ],
]


@pytest.fixture
def mocked_api(mocker):
    api_mock = mocker.Mock()
    api_mock.output_folder = "everest_output"
    api_mock.control_values = controls[0]
    api_mock.single_objective_values = objectives[0]
    return api_mock


def test_control_data_per_batch(mocker):
    api_mock = mocker.Mock()
    control_values = [
        {"control": "c1", "batch": 0, "value": 0},
        {"control": "c2", "batch": 0, "value": 0},
        {"control": "c1", "batch": 1, "value": 42},
        {"control": "c1", "batch": 1, "value": 24},
    ]
    api_mock.control_values = control_values
    result = _control_data_per_batch(api_mock)
    expected = pd.DataFrame(control_values)

    assert expected.equals(result)


@pytest.mark.parametrize(
    "controls, objectives, expected",
    [
        (controls[0], objectives[0], expected[0]),
        (controls[1], objectives[1], expected[1]),
        (controls[2], objectives[2], expected[2]),
    ],
)
def test_control_data_initial_vs_best(controls, objectives, expected, mocker):
    api_mock = mocker.Mock()
    api_mock.control_values = controls
    api_mock.single_objective_values = objectives
    result = _control_data_initial_vs_best(api_mock)
    expected_result = pd.DataFrame(expected)
    assert np.array_equal(result.values, expected_result.values)


def test_set_up_data_sources(tmpdir, mocked_api):
    expected = DataSources(
        controls_per_batch="everest_output/everviz/controls_per_batch.csv",
        controls_initial_vs_best="everest_output/everviz/controls_initial_vs_best.csv",
    )

    with tmpdir.as_cwd():
        for field in expected._fields:
            assert not os.path.exists(getattr(expected, field))

        get_everviz_folder(mocked_api)
        result = _set_up_data_sources(mocked_api)
        assert expected == result

        for field in result._fields:
            assert os.path.exists(getattr(result, field))



import subprocess
import shutil
import signal
import sys
import os

from everest.plugins import hookimpl  # pylint: disable=import-error
from everviz.util import get_everviz_folder
from everviz.config import setup_default_everviz_config
from everviz.log import setup_logger


def handle_exit(*args):  # pylint: disable=unused-argument)
    print("\n" + "=" * 32)
    print("Session terminated by the user.\nThank you for using Everviz!")
    print("=" * 32)
    sys.tracebacklimit = 0
    sys.stdout = open(  # pylint: disable=consider-using-with
        os.devnull, "w", encoding="utf-8"
    )
    sys.exit()


@hookimpl
def visualize_data(api):
    everviz_folder = get_everviz_folder(api)
    logger = setup_logger(everviz_folder)
    signal.signal(signal.SIGINT, handle_exit)

    config_path = setup_default_everviz_config(api)
    # The entry point of webviz is to call it from command line, and so do we.
    if shutil.which("webviz"):
        subprocess.call(["webviz", "build", config_path, "--theme", "equinor"])
    else:
        logger.error("Failed to find webviz")



import pandas as pd
from everviz.plugins.objectives_plot.objectives_plot import ObjectivesPlot
from everviz.pages.objectives import _objective_values


def test_objective_plot_callback(app, dash_duo, mocker, caplog):
    test_data = [
        {
            "batch": 0,
            "realization": 1,
            "function": "f0",
            "value": 100,
            "simulation": 1,
            "weight": 0.75,
            "norm": 1.5,
        },
        {
            "batch": 0,
            "realization": 1,
            "function": "f1",
            "value": 200,
            "simulation": 1,
            "weight": 0.25,
            "norm": 1,
        },
        {
            "batch": 2,
            "realization": 1,
            "function": "f0",
            "value": 300,
            "simulation": 1,
            "weight": 0.75,
            "norm": 1.5,
        },
        {
            "batch": 2,
            "realization": 1,
            "function": "f1",
            "value": 400,
            "simulation": 1,
            "weight": 0.25,
            "norm": 1,
        },
        {
            "batch": 0,
            "realization": 2,
            "function": "f0",
            "value": 500,
            "simulation": 2,
            "weight": 0.75,
            "norm": 1.5,
        },
        {
            "batch": 0,
            "realization": 2,
            "function": "f1",
            "value": 600,
            "simulation": 2,
            "weight": 0.25,
            "norm": 1,
        },
        {
            "batch": 2,
            "realization": 2,
            "function": "f0",
            "value": 700,
            "simulation": 2,
            "weight": 0.75,
            "norm": 1.5,
        },
        {
            "batch": 2,
            "realization": 2,
            "function": "f1",
            "value": 800,
            "simulation": 2,
            "weight": 0.25,
            "norm": 1,
        },
    ]

    mocker.patch(
        "everviz.plugins.objectives_plot.objectives_plot.get_data",
        return_value=_objective_values(pd.DataFrame(test_data)),
    )

    plugin = ObjectivesPlot(app, "values")
    app.layout = plugin.layout
    dash_duo.start_server(app)

    # Tests warning label not there
    dash_duo.wait_for_contains_text(f"#{plugin.label_id}", "")
    # Test statistics plot.
    dash_duo.find_element("#{} label:nth-child({})".format(plugin.radio_id, 1)).click()
    # Tests warning label is there
    dash_duo.wait_for_contains_text(f"#{plugin.label_id}", "Statistics are calculated")

    dash_duo.select_dcc_dropdown("#{}".format(plugin.function_dropdown_id), "f1")

    # Test values plot.
    dash_duo.find_element("#{} label:nth-child({})".format(plugin.radio_id, 2)).click()
    # Tests warning label not there
    dash_duo.wait_for_contains_text(f"#{plugin.label_id}", "")

    # Test normalized values
    dash_duo.find_element("#{} label:nth-child({})".format(plugin.radio_id, 3)).click()

    # Test normalized + weighted values
    dash_duo.find_element("#{} label:nth-child({})".format(plugin.radio_id, 4)).click()

    # Test plot mode
    dash_duo.find_element(
        "#{} label:nth-child({})".format(plugin.radio_id_mode, 1)
    ).click()
    dash_duo.find_element(
        "#{} label:nth-child({})".format(plugin.radio_id_mode, 2)
    ).click()

    # Test filtering realizations.
    dash_duo.find_element(
        "#{} label:nth-child({})".format(plugin.realization_filter_check_id, 1)
    ).click()
    dash_duo.find_element("#{}".format(plugin.realization_filter_input_id)).send_keys(
        "1, 3"
    )

    # Clear the dropdown, which should not cause an error.
    dash_duo.clear_input("#{}".format(plugin.function_dropdown_id))
    dash_duo.clear_input("#{}".format(plugin.function_dropdown_id))

    for record in caplog.records:
        assert record.levelname != "ERROR"



import os

import pytest
import pandas as pd

from everviz.util import get_everviz_folder
from everviz.pages.gradientplot import (
    _gradient_values,
    DataSources,
    _set_up_data_sources,
)


_DATA = [
    {"batch": 0, "function": "f1", "control": "c1", "value": 1},
    {"batch": 0, "function": "f1", "control": "c2", "value": 2},
    {"batch": 0, "function": "f2", "control": "c1", "value": 3},
    {"batch": 0, "function": "f2", "control": "c2", "value": 4},
    {"batch": 1, "function": "f1", "control": "c1", "value": 5},
    {"batch": 1, "function": "f1", "control": "c2", "value": 6},
    {"batch": 1, "function": "f2", "control": "c1", "value": 7},
    {"batch": 1, "function": "f2", "control": "c2", "value": 8},
]


@pytest.fixture
def mocked_api(mocker):
    api_mock = mocker.Mock()
    api_mock.output_folder = "everest_output"
    api_mock.gradient_values = _DATA
    return api_mock


def test_no_gradient_data(mocker):
    api_mock = mocker.Mock()
    api_mock.gradient_values = {}
    result = _gradient_values(api_mock)
    assert result is None


def test_gradient_data(mocker):
    api_mock = mocker.Mock()
    api_mock.gradient_values = _DATA
    result = _gradient_values(api_mock)
    expected = pd.DataFrame(_DATA)

    assert expected.equals(result)


def test_set_up_gradient_data_sources(tmpdir, mocked_api):
    expected = DataSources(
        gradient_values="everest_output/everviz/gradient_values.csv",
    )

    with tmpdir.as_cwd():
        for field in expected._fields:
            assert not os.path.exists(getattr(expected, field))

        get_everviz_folder(mocked_api)
        result = _set_up_data_sources(mocked_api)
        assert expected == result

        for field in result._fields:
            assert os.path.exists(getattr(result, field))



from functools import partial

import pandas as pd
import numpy as np


def calculate_statistics(data):
    data = data.reset_index()
    index = ["function", "batch"]
    result = pd.pivot_table(
        data,
        values="value",
        index=index,
        aggfunc=[
            np.mean,
            partial(np.quantile, q=0.1),
            partial(np.quantile, q=0.9),
            np.min,
            np.max,
        ],
    )
    min_index = data.groupby(index)["value"].idxmin()
    max_index = data.groupby(index)["value"].idxmax()
    min_realization = data.loc[min_index].set_index(index)["realization"]
    max_realization = data.loc[max_index].set_index(index)["realization"]
    result = pd.concat([result, min_realization, max_realization], axis=1)
    result.columns = [
        "mean",
        "P10",
        "P90",
        "min_value",
        "max_value",
        "min_realization",
        "max_realization",
    ]
    return result.sort_index().reset_index()



import os
from collections import namedtuple

from everviz.log import get_logger


DataSources = namedtuple("DataSource", ["summary_values"])

logger = get_logger()


def _summary_values(summary_values):
    # Sort by the index columns.
    sorted_values = (
        summary_values.drop(columns="simulation")
        .set_index(["batch", "date", "realization"])
        .dropna(axis=1, how="all")
        .sort_index()
        .reset_index()
    )
    return sorted_values


def _set_up_data_sources(api, keys=None):
    everest_folder = api.output_folder
    everviz_path = os.path.join(everest_folder, "everviz")

    # Make a table which statistics over the realizations.
    summary_values = api.summary_values(keys=keys)

    if summary_values.empty:
        return None

    logger.info("Generating summary values plot")

    summary_values_file = os.path.join(everviz_path, "summary_values.csv")
    values = _summary_values(summary_values)
    values.to_csv(summary_values_file, index=False)
    logger.info(f"File created: {summary_values_file}")  # pylint: disable=W1203

    return DataSources(summary_values=summary_values_file)


def page_layout(api):
    sources = _set_up_data_sources(api)

    if sources is None:
        return {}

    return {
        "title": "Summary Values",
        "content": [
            "## Summary values as a function of date",
            {
                "SummaryPlot": {
                    "csv_file": sources.summary_values,
                    "xaxis": "date",
                },
            },
            "## Summary values as a function of batch",
            {
                "SummaryPlot": {
                    "csv_file": sources.summary_values,
                    "xaxis": "batch",
                },
            },
        ],
    }



from uuid import uuid4

from dash import html, dcc
from dash.dependencies import Output, Input, State

from everviz.plugins.plugin_abc import EvervizPluginABC
from everviz.data.load_csv.get_data import get_data
from everviz.util import parse_range, get_placeholder_text, base64encode
from everviz.plugins.summary_plot import summary_callback

from everviz.plugins.summary_plot.util import calculate_statistics
from everviz.plugins.utils.layout.sidebar_layout import get_sidebar_layout


class SummaryPlot(EvervizPluginABC):
    """
    The SummaryPlot class implements a plugin for Webviz, for plotting summary
    key values generated by the eclipse or flow simulators during an Everest
    optimization.

    Summary values can be plotted as a function of time, or as a function of the
    batch number. Generally there are multiple realizations at each batch or
    date. These can be plotted either all individually in the plot, or they are
    summarized by plotting the mean value together with a P10-P90 range.
    """

    def __init__(self, app, csv_file, xaxis="date"):
        super().__init__()

        self.graph_id = f"graph-{uuid4()}"
        self.key1_dropdown_id = f"dropdown-{uuid4()}"
        self.key2_dropdown_id = f"dropdown-{uuid4()}"
        self.xaxis_dropdown_id = f"dropdown-{uuid4()}"
        self.radio_id = f"radio-{uuid4()}"
        self.realization_filter_check_id = f"check-{uuid4()}"
        self.realization_filter_input_id = f"input-{uuid4()}"
        self.label_id = f"label-{uuid4()}"

        self.csv_file = csv_file
        self.xaxis = xaxis

        self.set_callbacks(app)

    def add_webvizstore(self):
        return [
            (get_data, [{"csv_file": self.csv_file}]),
        ]

    @property
    def layout(self):
        radio_options = ["Statistics", "Data"]
        data = get_data(self.csv_file).set_index(["batch", "date", "realization"])
        key_dropdown_options = list(data.columns.unique())
        data = data.reset_index()
        realizations = data["realization"].unique()
        placeholder_text = get_placeholder_text(realizations)
        xaxis_dropdown_options = list(
            data["batch" if self.xaxis == "date" else "date"].unique()
        )
        xaxis_dropdown_title = (
            "Batches to plot" if self.xaxis == "date" else "Dates to plot"
        )

        side_bar_config = [
            (
                "dropdown",
                {
                    "title": "Keywords to plot:",
                    "item_id": self.key1_dropdown_id,
                    "options": key_dropdown_options,
                    "multi": False,
                },
            ),
            (
                "dropdown",
                {
                    "item_id": self.key2_dropdown_id,
                    "options": key_dropdown_options,
                    "multi": False,
                    "set_initial_value": False,
                },
            ),
            (
                "dropdown",
                {
                    "title": xaxis_dropdown_title,
                    "item_id": self.xaxis_dropdown_id,
                    "options": xaxis_dropdown_options,
                    "multi": True,
                },
            ),
            (
                "label",
                {
                    "item_id": self.label_id,
                    "text": f"Statistics are calculated based on {len(realizations)} realizations",
                    "style": {},
                },
            ),
            (
                "radio",
                {"item_id": self.radio_id, "options": radio_options},
            ),
        ]

        realization_elements = [
            html.Div(
                [
                    dcc.Checklist(
                        id=self.realization_filter_check_id,
                        options=[{"label": "Filter realizations:", "value": "filter"}],
                        style={"display": "inline-block", "margin-right": 8},
                    ),
                    dcc.Input(
                        id=self.realization_filter_input_id,
                        type="text",
                        placeholder=placeholder_text,
                        pattern=r"\s*|([0-9]+(\s*-\s*[0-9]+)?)(\s*,\s*[0-9]+(\s*-\s*[0-9]+)?)*",
                        style={"display": "inline-block"},
                    ),
                ],
                style={"margin-top": 20},
            )
        ]

        return html.Div(
            [
                html.Div(
                    [
                        html.Div(
                            [get_sidebar_layout(side_bar_config)]
                            + realization_elements,
                            style={
                                "width": "29%",
                                "display": "inline-block",
                                "vertical-align": "top",
                            },
                        ),
                        html.Div(
                            [
                                dcc.Graph(
                                    id=self.graph_id,
                                    config={
                                        "displaylogo": False,
                                        "toImageButtonOptions": {
                                            "filename": f"summary_by_{self.xaxis}"
                                        },
                                    },
                                )
                            ],
                            style={"width": "69%", "display": "inline-block"},
                        ),
                    ]
                ),
            ]
        )

    def set_callbacks(self, app):
        @app.callback(
            self.plugin_data_output,
            [self.plugin_data_requested],
            [
                State(self.key1_dropdown_id, "value"),
                State(self.key2_dropdown_id, "value"),
                State(self.radio_id, "value"),
                State(self.realization_filter_check_id, "value"),
                State(self.realization_filter_input_id, "value"),
            ],
        )
        def user_download_data(
            data_requested,
            key1,
            key2,
            radio_value,
            realizations_check,
            realizations_input,
        ):
            key_list = []
            if key1:
                key_list.append(key1)
            if key2:
                key_list.append(key2)

            if data_requested and key_list is not None and len(key_list) > 0:
                content = get_data(self.csv_file)
                if realizations_check:
                    realizations = parse_range(realizations_input)
                    if realizations:
                        content = content[content["realization"].isin(realizations)]
                if radio_value == "Statistics":
                    filename = "summary_statistics.csv"
                    content = calculate_statistics(content, key_list)
                else:
                    filename = "summary_values.csv"
                return {
                    "filename": filename,
                    "content": base64encode(content.to_csv()),
                    "mime_type": "text/csv",
                }
            return None

        @app.callback(
            Output(self.label_id, "style"),
            [Input(self.radio_id, "value")],
        )
        def radio_group_click(option):
            if option == "Statistics":
                return {}
            return {"display": "none"}

        @app.callback(
            Output(self.realization_filter_input_id, "disabled"),
            [Input(self.realization_filter_check_id, "value")],
        )
        def set_button_enabled_state(filter_realizations):
            return not filter_realizations

        # pylint: disable=too-many-locals
        @app.callback(
            Output(self.graph_id, "figure"),
            [
                Input(self.key1_dropdown_id, "value"),
                Input(self.key2_dropdown_id, "value"),
                Input(self.xaxis_dropdown_id, "value"),
                Input(self.radio_id, "value"),
                Input(self.realization_filter_check_id, "value"),
                Input(self.realization_filter_input_id, "value"),
            ],
        )
        def update_graph(
            key1, key2, line_list, radio_value, realizations_check, realizations_input
        ):
            # The key_list arguments is the list of keys to plot. The line_list
            # argument is a list of batches, or a list of dates to plot for
            # those keys.
            key_list = []
            if key1:
                key_list.append(key1)
            if key2:
                key_list.append(key2)

            if not key_list or not line_list:
                return {}

            data = get_data(self.csv_file)
            if realizations_check:
                realizations = parse_range(realizations_input)
                if realizations:
                    data = data[data["realization"].isin(realizations)]

            if radio_value == "Statistics":
                data = calculate_statistics(data, key_list).set_index(
                    ["summary_key", "batch", "date"]
                )
            else:
                data = data.set_index(["batch", "date"])

            callback = summary_callback.get_callback_func(radio_value)

            line_key = "batch" if self.xaxis == "date" else "date"
            line_filter = "batch" if line_key == "date" else "date"
            traces = callback(data, key_list, line_list, line_key, line_filter)
            return {
                "data": traces,
                "layout": summary_callback.get_layout(
                    summary_key_list=key_list, xaxis_title=self.xaxis.capitalize()
                ),
            }



import os
from collections import namedtuple
import pandas as pd
from everviz.log import get_logger


DataSources = namedtuple(
    "DataSource", ["objective_delta_values", "summary_delta_values"]
)

logger = get_logger()


def _get_objective_delta_values(api, best_batch):
    data = pd.DataFrame(api.objective_values).drop(columns=["simulation"])
    initial = data[(data.batch == 0)].drop(columns=["batch"])
    best = data[data.batch == best_batch].drop(columns=["batch"])
    initial = initial.pivot(index="realization", values="value", columns="function")
    best = best.pivot(index="realization", values="value", columns="function")
    return (best - initial).reset_index()


def _get_summary_delta_values(api, best_batch):
    summary = api.summary_values()
    if summary.empty:
        return summary
    data = (
        summary.drop(columns="simulation")
        .set_index(["realization", "date"])
        .dropna(axis=1, how="all")
    )
    if not data.empty:
        initial = data[(data.batch == 0)].drop(columns=["batch"])
        best = data[data.batch == best_batch].drop(columns=["batch"])
        return (best - initial).reset_index()
    return data


def _set_up_data_sources(api):
    everest_folder = api.output_folder
    everviz_path = os.path.join(everest_folder, "everviz")

    objectives_df = pd.DataFrame(api.single_objective_values)
    best_batch = objectives_df.batch[objectives_df.objective.idxmax()]

    logger.info("Generating objective delta plot source data file")
    objective_csv_path = os.path.join(everviz_path, "objective_delta_values.csv")
    data = _get_objective_delta_values(api, best_batch)
    data.to_csv(objective_csv_path, index=False)
    logger.info(f"File created: {objective_csv_path}")  # pylint: disable=W1203

    logger.info("Generating summary delta plot source data file")
    summary_csv_path = os.path.join(everviz_path, "summary_delta_values.csv")
    data = _get_summary_delta_values(api, best_batch)
    if not data.empty:
        data.to_csv(summary_csv_path, index=False)
        logger.info(f"File created: {summary_csv_path}")  # pylint: disable=W1203
    else:
        logger.info("No summary data, not creating summary_delta_values.csv")
        summary_csv_path = None

    return DataSources(
        objective_delta_values=objective_csv_path, summary_delta_values=summary_csv_path
    )


def page_layout(api):
    sources = _set_up_data_sources(api)
    has_summary = sources.summary_delta_values is not None
    return {
        "title": "Objectives Delta Values",
        "content": [
            "## Objective functions: Difference between best and initial batch",
            {
                "DeltaPlot": {
                    "csv_file": sources.objective_delta_values,
                    "pre_select": "all",
                }
            },
        ]
        + (
            [
                "## Summary keys: Difference between best and initial batch",
                {
                    "DeltaPlot": {
                        "csv_file": sources.summary_delta_values,
                        "pre_select": "none",
                    },
                },
            ]
            if has_summary
            else ["## Summary keys: No data"]
        ),
    }



from uuid import uuid4
from pathlib import Path

import plotly.graph_objs as go

from dash import html, dcc
from dash.dependencies import Output, Input
from everviz.plugins.plugin_abc import EvervizPluginABC
from everviz.data.load_csv.get_data import get_data
from everviz.util import base64encode


class DeltaPlot(EvervizPluginABC):
    """
    The DeltaPlot class implements a plugin for Webviz, for plotting the
    difference in values generated during an Everest optimization.
    """

    def __init__(self, app, csv_file, pre_select):
        super().__init__()

        self.graph_id = f"graph-{uuid4()}"
        self.key_dropdown_id = f"dropdown-{uuid4()}"
        self.date_dropdown_id = f"dropdown-{uuid4()}"

        self.csv_file = csv_file
        self.pre_select = pre_select

        self.set_callbacks(app)

    def add_webvizstore(self):
        return [
            (get_data, [{"csv_file": self.csv_file}]),
        ]

    @property
    def layout(self):
        data = get_data(self.csv_file)

        key_dropdown_options = [
            {"label": i, "value": i}
            for i in data.columns
            if i not in ["realization", "date"]
        ]

        if self.pre_select == "first":
            key_dropdown_value = [key_dropdown_options[0]["label"]]
        elif self.pre_select == "all":
            key_dropdown_value = [option["label"] for option in key_dropdown_options]
        elif self.pre_select == "none":
            key_dropdown_value = []
        else:
            raise RuntimeError(f"Invalid argument: 'pre_select={self.pre_select}")

        dropdown_elements = [
            html.Label("Values to plot:"),
            dcc.Dropdown(
                id=self.key_dropdown_id,
                options=key_dropdown_options,
                multi=True,
                value=key_dropdown_value,
            ),
        ]

        if "date" in data.columns:
            date_dropdown_options = [
                {"label": i, "value": i} for i in data["date"].unique()
            ]
            date_dropdown_value = (
                [date_dropdown_options[-1]["label"]] if date_dropdown_options else None
            )
            dropdown_elements.extend(
                [
                    html.Label("Dates to plot:"),
                    dcc.Dropdown(
                        id=self.date_dropdown_id,
                        options=date_dropdown_options,
                        multi=True,
                        value=date_dropdown_value,
                    ),
                ]
            )

        if "objective" in self.csv_file:
            plot_filename = "delta_objectives"
        elif "summary" in self.csv_file:
            plot_filename = "delta_summary"
        else:
            plot_filename = "delta"

        return html.Div(
            [
                html.Div(
                    [
                        html.Div(
                            dropdown_elements,
                            style={
                                "width": "29%",
                                "display": "inline-block",
                                "vertical-align": "top",
                            },
                        ),
                        html.Div(
                            [
                                dcc.Graph(
                                    id=self.graph_id,
                                    config={
                                        "displaylogo": False,
                                        "toImageButtonOptions": {
                                            "filename": plot_filename
                                        },
                                    },
                                )
                            ],
                            style={"width": "69%", "display": "inline-block"},
                        ),
                    ]
                ),
            ]
        )

    def set_callbacks(self, app):
        @app.callback(self.plugin_data_output, [self.plugin_data_requested])
        def user_download_data(data_requested):
            if data_requested:
                return {
                    "filename": Path(self.csv_file).name,
                    "content": base64encode(csv=get_data(self.csv_file).to_csv()),
                    "mime_type": "text/csv",
                }
            return None

        inputs = [Input(self.key_dropdown_id, "value")]
        data = get_data(self.csv_file)
        if "date" in data.columns:
            inputs.append(Input(self.date_dropdown_id, "value"))

        @app.callback(
            Output(self.graph_id, "figure"),
            inputs,
        )
        def update_graph(key_list, *args):
            if key_list is None:
                return {}

            data = get_data(self.csv_file)

            traces = []
            if "date" in data.columns:
                for key in key_list:
                    for date in args[0]:
                        selected_data = data[data["date"] == date].drop(
                            columns=["date"]
                        )
                        traces.append(
                            go.Bar(
                                y=selected_data[key],
                                x=selected_data["realization"],
                                name=f"{key}:{date}",
                                showlegend=True,
                            ),
                        )
            else:
                for key in key_list:
                    traces.append(
                        go.Bar(
                            y=data[key],
                            x=data["realization"],
                            name=key,
                            showlegend=True,
                        ),
                    )

            return {
                "data": traces,
                "layout": dict(
                    xaxis={"title": "Realization", "tickformat": ",d"},
                    yaxis={"title": "(best realization) - (initial realization)"},
                ),
            }



from datetime import datetime
import pandas as pd
from everviz.plugins.summary_plot.summary_plot import SummaryPlot
from everviz.pages.summary_values import _summary_values


def test_summary_plot_callback(app, dash_duo, mocker, caplog):
    test_data = {
        "realization": range(12),
        "simulation": range(12),
        "batch": [0] * 6 + [1] * 6,
        "date": [
            datetime(2000, 1, 1),
            datetime(2000, 2, 1),
            datetime(2000, 3, 1),
            datetime(2000, 1, 1),
            datetime(2000, 2, 1),
            datetime(2000, 3, 1),
            datetime(2000, 1, 1),
            datetime(2000, 2, 1),
            datetime(2000, 3, 1),
            datetime(2000, 1, 1),
            datetime(2000, 2, 1),
            datetime(2000, 3, 1),
        ],
        "key1": range(1, 13),
        "key2": range(10, 130, 10),
    }

    mocker.patch(
        "everviz.plugins.summary_plot.summary_plot.get_data",
        return_value=_summary_values(pd.DataFrame(test_data)),
    )

    plugin = SummaryPlot(app, "values")
    app.layout = plugin.layout
    dash_duo.start_server(app)

    # Test statistics plot.
    dash_duo.wait_for_contains_text(f"#{plugin.label_id}", "Statistics are calculated")

    # Test data plot.
    dash_duo.find_element("#{} label:nth-child({})".format(plugin.radio_id, 2)).click()
    dash_duo.clear_input("#{}".format(plugin.key1_dropdown_id))
    dash_duo.select_dcc_dropdown("#{}".format(plugin.key1_dropdown_id), "key1")

    # Tests warning label not there
    dash_duo.wait_for_contains_text(f"#{plugin.label_id}", "")

    # Test filtering realizations.
    dash_duo.find_element(
        "#{} label:nth-child({})".format(plugin.realization_filter_check_id, 1)
    ).click()
    dash_duo.find_element("#{}".format(plugin.realization_filter_input_id)).send_keys(
        "1, 3"
    )

    # Clear the axes dropdowns, which should not cause an error.
    #
    # NOTE: if you try to do this before the previous click action on a radio
    # button, you get an error, because the the dropdown stays open and overlaps
    # with the radio buttons.
    dash_duo.clear_input("#{}".format(plugin.xaxis_dropdown_id))
    dash_duo.clear_input("#{}".format(plugin.key1_dropdown_id))

    for record in caplog.records:
        assert record.levelname != "ERROR"



from itertools import cycle
from datetime import datetime
from unittest import mock

import pytest
import pandas as pd
from plotly.colors import DEFAULT_PLOTLY_COLORS

from everviz.plugins.summary_plot import summary_callback
from everviz.plugins.summary_plot.util import calculate_statistics

__TEST_DATA = {
    "realization": range(12),
    "batch": [0] * 6 + [1] * 6,
    "date": [
        datetime(2000, 1, 1),
        datetime(2000, 2, 1),
        datetime(2000, 3, 1),
        datetime(2000, 1, 1),
        datetime(2000, 2, 1),
        datetime(2000, 3, 1),
        datetime(2000, 1, 1),
        datetime(2000, 2, 1),
        datetime(2000, 3, 1),
        datetime(2000, 1, 1),
        datetime(2000, 2, 1),
        datetime(2000, 3, 1),
    ],
    "key1": range(1, 13),
    "key2": range(10, 130, 10),
}


@pytest.mark.parametrize(
    "lines, x_filter, x_key",
    [([datetime(2000, 1, 1)], "date", "batch"), ([0], "batch", "date")],
)
def test_get_data_callback(monkeypatch, lines, x_filter, x_key):
    plotly_mock = mock.Mock()
    plotly_mock.Scattergl.return_value = "called_scatter"
    monkeypatch.setattr(summary_callback, "go", plotly_mock)
    df = pd.DataFrame(__TEST_DATA)
    df = df.set_index(["batch", "date"])
    summary_callback._get_data_lines(df, ["key1"], lines, x_filter, x_key)
    plotly_mock.Scattergl.assert_called()


@pytest.mark.parametrize(
    "keys, lines, x_filter, x_key",
    [
        ([], [datetime(2000, 1, 1)], "date", "batch"),
        (None, [0], "date", "batch"),
        (["key1"], [datetime(2000, 1, 1)], "date", "batch"),
        (["key1"], [0], "batch", "date"),
        (["key1", "key2"], [datetime(2000, 1, 1)], "date", "batch"),
        (["key1", "key2"], [0], "batch", "date"),
    ],
)
def test_get_lines_callback(monkeypatch, keys, lines, x_filter, x_key):
    plotly_mock = mock.Mock()
    plotly_mock.Scattergl.return_value = "scatter"
    monkeypatch.setattr(summary_callback, "go", plotly_mock)
    df = pd.DataFrame(__TEST_DATA)

    try:
        df = calculate_statistics(df, keys).set_index(["summary_key", "batch", "date"])

        traces = summary_callback._get_statistics_lines(
            df, ["key1"], lines, x_filter, x_key
        )

        assert plotly_mock.Scattergl.call_count == 5
        assert traces == 5 * ["scatter"]
    except AssertionError as e:
        assert str(e) in "List of summary keys is required"


@pytest.mark.parametrize(
    "keys, x_axis_title",
    [
        ([], "date"),
        (["key1", "key2", "key3"], "batch"),
    ],
)
def test_get_layout(keys, x_axis_title):
    expected = dict(
        xaxis={"title": x_axis_title},
        hovermode="closest",
    )
    colors = cycle(DEFAULT_PLOTLY_COLORS)
    for idx, key in enumerate(keys):
        color = next(colors)
        y_axis = f"yaxis{idx + 1}" if idx > 0 else "yaxis"
        expected.update(
            {
                y_axis: dict(
                    title=key,
                    titlefont=dict(color=color),
                    tickfont=dict(color=color),
                )
            }
        )
        if idx > 0:
            expected[y_axis].update(
                dict(
                    anchor="x",
                    overlaying="y",
                    side="right" if idx % 2 else "left",
                )
            )

    layout = summary_callback.get_layout(keys, x_axis_title)
    assert expected == layout
    if len(keys) > 0:
        assert "yaxis1" not in layout
        assert "yaxis3" in layout



import pandas as pd
from selenium.webdriver.common.by import By

import everviz
from everviz.plugins.crossplot.crossplot_indexed import CrossplotIndexed


def test_crossplot_indexed_callback(app, dash_duo, monkeypatch, mocker, caplog):
    columns_names = ["a", "b", "c"]
    mock_data = pd.DataFrame(
        data=[[1, 2, 3]],
        columns=[f"{name}-{idx}" for idx, name in enumerate(columns_names)],
    )
    monkeypatch.setattr(
        everviz.plugins.crossplot.crossplot_indexed,
        "get_data",
        mocker.Mock(return_value=mock_data),
    )

    plugin = CrossplotIndexed(app, "a_data_file")
    app.layout = plugin.layout
    dash_duo.start_server(app)

    dropdown_x = dash_duo.find_element(f"#{plugin.dropdown_x_id}")
    dropdown_x.click()
    options_x = dropdown_x.find_element(By.CSS_SELECTOR, "div.Select-menu-outer")
    assert columns_names == options_x.text.split("\n")
    dash_duo.clear_input("#{}".format(plugin.dropdown_x_id))

    dash_duo.click_at_coord_fractions(
        dash_duo.find_element(f"#{plugin.graph_id}"), 0, 0
    )

    dropdown_y = dash_duo.find_element(f"#{plugin.dropdown_y_id}")
    dropdown_y.click()
    options_y = dropdown_y.find_element(By.CSS_SELECTOR, "div.Select-menu-outer")
    assert columns_names == options_y.text.split("\n")
    dash_duo.clear_input("#{}".format(plugin.dropdown_y_id))

    dash_duo.clear_input("#{}".format(plugin.dropdown_realization_id))

    for record in caplog.records:
        assert record.levelname != "ERROR"



import os
from collections import namedtuple
import pandas as pd
from everviz.log import get_logger


DataSources = namedtuple("DataSource", "gradient_values")

logger = get_logger()


def _gradient_values(api):
    gradient_values = api.gradient_values
    if gradient_values:
        return pd.DataFrame(api.gradient_values)
    return None


def _set_up_data_sources(api):
    everest_folder = api.output_folder
    everviz_path = os.path.join(everest_folder, "everviz")

    logger.info("Generating controls per batch source data file")
    gradient_values = os.path.join(everviz_path, "gradient_values.csv")
    data = _gradient_values(api)
    if data is None:
        logger.info("No gradient data availabe, skipping")
        return None
    data.to_csv(gradient_values, index=False)
    logger.info(f"File created: {gradient_values}")  # pylint: disable=W1203
    return DataSources(gradient_values=gradient_values)


def page_layout(api):
    sources = _set_up_data_sources(api)
    if sources is None:
        return {}
    return {
        "title": "Gradients",
        "content": [
            "## Gradient values",
            {"GradientPlot": {"csv_file": sources.gradient_values}},
        ],
    }



from dash import dcc
from dash import html


def get_sidebar_layout(configurations):
    _FUNCTION_MAP = {
        "radio": _get_radio,
        "dropdown": _get_dropdown,
        "label": _get_label,
    }

    divs = []
    for comp_name, args in configurations:
        divs += _FUNCTION_MAP[comp_name](**args)
    return html.Div(divs)


def _get_radio(item_id, title=None, options=None):
    if not options:
        options = []
    try:
        options = [{"label": i, "value": j} for i, j in options]
    except ValueError:
        options = [{"label": i, "value": i} for i in options]
    return [
        title,
        dcc.RadioItems(
            id=item_id,
            options=options,
            value=options[0]["value"] if len(options) > 0 else None,
            labelStyle={"display": "inline-block"},
        ),
    ]


def _get_dropdown(
    item_id, title=None, options=None, multi=False, set_initial_value=True
):
    if options is None:
        options = []
    options = [{"label": i, "value": i} for i in options]
    if set_initial_value:
        initial_value = options[0]["value"] if len(options) > 0 else []
    else:
        initial_value = None
    return [
        title,
        dcc.Dropdown(
            id=item_id,
            options=options,
            placeholder="Select a label",
            value=[initial_value] if multi else initial_value,
            multi=multi,
        ),
    ]


def _get_label(item_id, text, style):
    return [html.Label(text, id=item_id, style=style)]



import os
import base64
from collections import defaultdict

DEFAULT_CONFIG = "everviz_webviz_config.yml"


def get_everviz_folder(api):
    everest_folder = api.output_folder
    everviz_path = os.path.join(everest_folder, "everviz")
    if not os.path.exists(everviz_path):
        os.makedirs(everviz_path)
    return everviz_path


def identify_indexed_controls(controls):
    base_names = defaultdict(list)
    for control in controls:
        try:
            base_name, index = control.split("-")
            if index.isnumeric():
                base_names[base_name].append(int(index))
        except ValueError:
            continue

    return base_names


def parse_range(numbers):
    result = set()
    if numbers is not None:
        for number in numbers.split(","):
            number = number.strip()
            if number.isdigit():
                result.add(int(number))
            elif "-" in number:
                number_range = number.split("-")
                start = number_range[0]
                stop = number_range[1]
                if start.isdigit() and stop.isdigit():
                    result |= set(range(int(start), int(stop) + 1))
    return result


def get_placeholder_text(realizations):
    if len(realizations) == 0:
        return "No realizations found"

    if len(realizations) > 1:
        min_r = min(realizations)
        max_r = max(realizations)
        return f"example: {min_r}, {max_r}, {min_r}-{max_r}"

    return f"example: {realizations[0]}"


def base64encode(csv):
    return base64.b64encode(csv.encode("ascii")).decode("ascii")



from uuid import uuid4
from pathlib import Path

import plotly.graph_objs as go

from dash import html, dcc
from dash.dependencies import Output, Input
from everviz.plugins.plugin_abc import EvervizPluginABC
from everviz.data.load_csv.get_data import get_data
from everviz.util import base64encode


class ControlsPlot(EvervizPluginABC):
    """
    The ControlsPlot class implements a plugin for Webviz, for plotting the
    control values generated during an Everest optimization.
    """

    def __init__(self, app, csv_file):
        super().__init__()

        self.graph_id = f"graph-{uuid4()}"
        self.controls_dropdown_id = f"dropdown-{uuid4()}"

        self.csv_file = csv_file
        self.set_callbacks(app)

    def add_webvizstore(self):
        return [(get_data, [{"csv_file": self.csv_file}])]

    @property
    def layout(self):
        data = get_data(self.csv_file)

        controls = data["control"].unique()
        control_dropdown_options = [{"label": i, "value": i} for i in controls]
        control_elements = [
            html.Label("Controls to plot:"),
            dcc.Dropdown(
                id=self.controls_dropdown_id,
                options=control_dropdown_options,
                multi=True,
                value=[controls[0]],
            ),
        ]

        return html.Div(
            [
                html.Div(
                    [
                        html.Div(
                            control_elements,
                            style={
                                "width": "29%",
                                "display": "inline-block",
                                "vertical-align": "top",
                            },
                        ),
                        html.Div(
                            [
                                dcc.Graph(
                                    id=self.graph_id,
                                    config={
                                        "displaylogo": False,
                                        "toImageButtonOptions": {
                                            "filename": "controls"
                                        },
                                    },
                                )
                            ],
                            style={"width": "69%", "display": "inline-block"},
                        ),
                    ]
                ),
            ]
        )

    def set_callbacks(self, app):
        @app.callback(self.plugin_data_output, [self.plugin_data_requested])
        def user_download_data(data_requested):
            if data_requested:
                return {
                    "filename": Path(self.csv_file).name,
                    "content": base64encode(get_data(self.csv_file).to_csv()),
                    "mime_type": "text/csv",
                }
            return None

        @app.callback(
            Output(self.graph_id, "figure"),
            [Input(self.controls_dropdown_id, "value")],
        )
        def update_graph(control_list):
            # # The key_list arguments is the list of functions to plot.
            if control_list is None:
                return {}

            # # Get the data, setting its index.
            data = get_data(self.csv_file).set_index("control")

            traces = []
            for control in control_list:
                control_data = data.xs(control)
                traces.append(
                    go.Scattergl(
                        y=control_data["value"],
                        x=control_data["batch"],
                        mode="lines",
                        name=control,
                        showlegend=True,
                    ),
                )
            return {
                "data": traces,
                "layout": dict(
                    xaxis={"title": "Batch", "tickformat": ",d"},
                    yaxis={"title": "Control Value"},
                ),
            }



import pandas as pd
import pytest
from everviz.plugins.crossplot.callback.crossplot_callback import (
    crossplot_update_graph,
    get_graph_line,
)


@pytest.mark.parametrize("axis", ["xaxis", "yaxis"])
@pytest.mark.parametrize("axis_type, expected", [("Log", "log"), ("Linear", "linear")])
def test_crossplot_callback_axis_type(axis, axis_type, expected):
    if axis == "xaxis":
        inactive_axis = "yaxis"
        test_result = crossplot_update_graph(
            [], "x-column", "y-column", axis_type, "Log"
        )
    else:
        inactive_axis = "xaxis"
        test_result = crossplot_update_graph(
            [], "x-column", "y-column", "Log", axis_type
        )

    assert list(test_result.keys()) == ["data", "layout"]

    assert test_result["layout"][axis]["type"] == expected
    assert test_result["layout"][inactive_axis]["type"] == "log"


@pytest.mark.parametrize("axis", ["x", "y"])
@pytest.mark.parametrize(
    "test_input, expected", [("an-axis-type", [1, 2]), ("Cumulative", [1, 3])]
)
def test_crossplot_get_trace(axis, test_input, expected):
    mock_data = pd.DataFrame(data=[[1, 1], [2, 2]], columns=["a", "b"])
    if axis == "x":
        result = get_graph_line(mock_data["a"], mock_data["b"], test_input, "")
    else:
        result = get_graph_line(mock_data["a"], mock_data["b"], "", test_input)

    assert all(result[axis] == expected)



import os
from datetime import datetime

import pandas

from everviz.pages.deltaplot import (
    _get_objective_delta_values,
    _get_summary_delta_values,
    _set_up_data_sources,
    page_layout,
)

_OBJECTIVES = [
    {"batch": 0, "realization": 1, "function": "f0", "value": 100, "simulation": 1},
    {"batch": 0, "realization": 2, "function": "f0", "value": 200, "simulation": 2},
    {"batch": 0, "realization": 1, "function": "f1", "value": 300, "simulation": 1},
    {"batch": 0, "realization": 2, "function": "f1", "value": 400, "simulation": 2},
    {"batch": 2, "realization": 1, "function": "f0", "value": 500, "simulation": 1},
    {"batch": 2, "realization": 2, "function": "f0", "value": 600, "simulation": 2},
    {"batch": 2, "realization": 1, "function": "f1", "value": 800, "simulation": 1},
    {"batch": 2, "realization": 2, "function": "f1", "value": 900, "simulation": 2},
]

_SUMMARY = {
    "realization": 3 * [1] + 3 * [2] + 3 * [1] + 3 * [2],
    "simulation": 3 * [1] + 3 * [2] + 3 * [1] + 3 * [2],
    "batch": [0] * 6 + [2] * 6,
    "date": 4
    * [
        datetime(2000, 1, 1).date().isoformat(),
        datetime(2000, 2, 1).date().isoformat(),
        datetime(2000, 3, 1).date().isoformat(),
    ],
    "key1": range(1, 13),
    "key2": range(10, 130, 10),
}

_SINGLE_OBJECTIVES = {"batch": [0, 2], "objective": [0.0, 1.0]}


def test_objective_values_data_frame(mocker):
    """Test for the correct layout and size of the objective values data frame."""
    mock_api = mocker.Mock()
    mock_api.objective_values = _OBJECTIVES

    objective_delta_values = _get_objective_delta_values(mock_api, 2)

    assert set(objective_delta_values.columns) == {"realization", "f0", "f1"}
    assert len(objective_delta_values) == 2
    assert set(objective_delta_values["realization"]) == {
        obj["realization"] for obj in _OBJECTIVES
    }
    assert (objective_delta_values["f0"] == 400).all()
    assert (objective_delta_values["f1"] == 500).all()


def test_summary_values_data_frame(mocker):
    """Test for the correct layout and size of the objective values data frame."""
    mock_api = mocker.Mock()
    mock_api.summary_values.return_value = pandas.DataFrame(_SUMMARY)

    summary_delta_values = _get_summary_delta_values(mock_api, 2)

    assert set(summary_delta_values.columns) == {"realization", "date", "key1", "key2"}

    assert len(summary_delta_values) == 6
    assert set(summary_delta_values["realization"]) == set(_SUMMARY["realization"])
    assert (summary_delta_values["key1"] == 6).all()
    assert (summary_delta_values["key2"] == 60).all()


def test_set_up_sources(mocker, tmpdir):
    mock_api = mocker.Mock()
    mock_api.objective_values = _OBJECTIVES
    mock_api.single_objective_values = _SINGLE_OBJECTIVES
    mock_api.summary_values.return_value = pandas.DataFrame(_SUMMARY)
    mock_api.output_folder = tmpdir

    os.mkdir(os.path.join(tmpdir, "everviz"))
    data_source = _set_up_data_sources(mock_api)
    assert (
        data_source.objective_delta_values
        == tmpdir / "everviz" / "objective_delta_values.csv"
    )
    assert (
        data_source.summary_delta_values
        == tmpdir / "everviz" / "summary_delta_values.csv"
    )


def test_delta_plot_layout_with_empty_summary(mocker, tmpdir):
    mock_api = mocker.Mock()
    mock_api.objective_values = _OBJECTIVES
    mock_api.single_objective_values = _SINGLE_OBJECTIVES
    mock_api.summary_values.return_value = pandas.DataFrame(
        {"realization": [], "simulation": [], "date": [], "batch": []}
    )
    mock_api.output_folder = tmpdir

    os.mkdir(os.path.join(tmpdir, "everviz"))
    layout = page_layout(mock_api)

    assert "content" in layout
    content = layout["content"]
    assert len(content) == 3
    assert "Objective functions" in content[0]
    assert "DeltaPlot" in content[1]
    assert "Summary keys: No data" in content[2]



from everviz.pages.objectives import _single_objective_title


def test_single_objective_title(mocker):
    api_mock = mocker.Mock()
    api_mock.objective_function_names = ["npv"]
    result = _single_objective_title(api_mock)
    expected = "Objective"
    assert result == expected

    api_mock.objective_function_names = ["npv", "rf"]
    result = _single_objective_title(api_mock)
    expected = "Objective functions"
    assert result == expected



import os
import yaml

from everviz.log import get_logger
from everviz.util import get_everviz_folder, DEFAULT_CONFIG
from everviz.pages import (
    controls,
    configuration,
    crossplot,
    summary_values,
    objectives,
    deltaplot,
    wells_values,
    gradientplot,
)

logger = get_logger()


def webviz_config(api):
    pages = [
        {
            "title": "Everest",
            "content": [],
        },
        objectives.page_layout(api),
        summary_values.page_layout(api),
        wells_values.page_layout(api),
        crossplot.page_layout(api),
        controls.page_layout(api),
        gradientplot.page_layout(api),
        deltaplot.page_layout(api),
        configuration.page_layout(api),
    ]

    # Remove possible empty pages
    pages = list(filter(None, pages))

    return {
        "title": "Everest Optimization Report",
        "pages": pages,
    }


def write_webviz_config(config, file_path):
    with open(file_path, "w", encoding="utf-8") as fh:
        yaml.dump(config, fh, default_flow_style=False, sort_keys=False)
    logger.info("Webviz config file created: {}".format(file_path))


def setup_default_everviz_config(api):
    """
    Creates default everviz config file and all required data sources
    """
    everviz_folder = get_everviz_folder(api)

    config = webviz_config(api)
    config_file_path = os.path.join(everviz_folder, DEFAULT_CONFIG)
    write_webviz_config(config, config_file_path)
    logger.info("Default everviz config created: {}".format(config_file_path))
    return config_file_path



import os
import pytest
from everviz.util import get_everviz_folder, parse_range, get_placeholder_text


@pytest.fixture
def mocked_api(mocker):
    api_mock = mocker.Mock()
    api_mock.output_folder = "everest_output"
    return api_mock


def test_get_everviz_folder(tmpdir, mocked_api):
    everest_folder = "everest_output"
    expected_everviz_path = os.path.join(everest_folder, "everviz")
    with tmpdir.as_cwd():
        assert not os.path.exists(expected_everviz_path)
        everviz_path = get_everviz_folder(mocked_api)
        assert os.path.exists(everviz_path)
        assert expected_everviz_path == everviz_path


@pytest.mark.parametrize(
    "input_string, expected",
    [
        ("", set()),
        (" ", set()),
        ("1,2,3", {1, 2, 3}),
        ("1, 2 ,3", {1, 2, 3}),
        ("3,1,2", {1, 2, 3}),
        ("3,1,3", {1, 3}),
        ("1-3", {1, 2, 3}),
        ("1, 2-3", {1, 2, 3}),
        ("1, 1-3, 2", {1, 2, 3}),
        ("x", set()),
        ("1,2,x,3", {1, 2, 3}),
    ],
)
def test_object_parse_range(input_string, expected):
    assert parse_range(input_string) == expected


@pytest.mark.parametrize(
    "realizations, expected",
    [
        ([], "No realizations found"),
        ([1, 2, 3, 5], "example: 1, 5, 1-5"),
        ([0], "example: 0"),
    ],
)
def test_get_placeholder_text(realizations, expected):
    assert get_placeholder_text(realizations) == expected



from uuid import uuid4
from pathlib import Path
from itertools import cycle

import plotly.graph_objs as go
from plotly.colors import DEFAULT_PLOTLY_COLORS

from dash import html, dcc
from dash.dependencies import Output, Input
from everviz.data.load_csv.get_data import get_data
from everviz.plugins.plugin_abc import EvervizPluginABC
from everviz.util import base64encode


class SingleObjectivesPlot(EvervizPluginABC):
    """
    This webviz plugin is designed to display the single objective function
    calculated as the weighted sum of the objective functions from a multi
    objective everest optimisation.
    """

    def __init__(self, app, csv_file):
        super().__init__()

        self.graph_id = f"graph-{uuid4()}"
        self.div_id = f"div-{uuid4()}"
        self.function_dropdown_id = f"dropdown-{uuid4()}"
        self.csv_file = csv_file
        self.set_callbacks(app)

    def add_webvizstore(self):
        return [
            (get_data, [{"csv_file": self.csv_file}]),
        ]

    @property
    def layout(self):
        data = get_data(self.csv_file)
        exclude = ["batch", "accepted"]
        functions = [name for name in data.columns if name not in exclude]
        function_dropdown_options = [{"label": i, "value": i} for i in functions]
        func_elements = [
            html.Label("Functions to plot:"),
            dcc.Dropdown(
                id=self.function_dropdown_id,
                options=function_dropdown_options,
                multi=True,
                value=[functions[0]],
            ),
        ]
        return html.Div(
            [
                html.Div(
                    func_elements,
                    style={
                        "width": "29%",
                        "display": "inline-block",
                        "vertical-align": "top",
                    },
                ),
                html.Div(
                    [
                        dcc.Graph(
                            id=self.graph_id,
                            config={
                                "displaylogo": False,
                                "toImageButtonOptions": {
                                    "filename": "single_objectives"
                                },
                            },
                        )
                    ],
                    id=self.div_id,
                    style={"width": "69%", "display": "inline-block"},
                ),
            ]
        )

    def set_callbacks(self, app):
        @app.callback(self.plugin_data_output, [self.plugin_data_requested])
        def user_download_data(data_requested):
            if data_requested:
                return {
                    "filename": Path(self.csv_file).name,
                    "content": base64encode(csv=get_data(self.csv_file).to_csv()),
                    "mime_type": "text/csv",
                }
            return None

        @app.callback(
            Output(self.graph_id, "figure"),
            [Input(self.function_dropdown_id, "value"), Input(self.div_id, "children")],
        )
        def update_graph(function_list, _):
            if function_list is None:
                return {}

            data = get_data(self.csv_file)
            accepted_data = data[data.accepted == 1]
            not_accepted_data = data[data.accepted == 0]

            colors = cycle(DEFAULT_PLOTLY_COLORS)
            traces = []
            for name, color in zip(function_list, colors):
                if name != "objective":
                    traces.append(
                        go.Scattergl(
                            y=data[name],
                            x=data.batch,
                            mode="lines+markers",
                            marker={"color": color, "size": 10},
                            name=name,
                        ),
                    )
                else:
                    traces.extend(
                        [
                            go.Scattergl(
                                y=accepted_data[name],
                                x=accepted_data.batch,
                                mode="lines+markers",
                                marker={"color": "blue", "size": 10},
                                name="accepted",
                            ),
                            go.Scattergl(
                                y=not_accepted_data[name],
                                x=not_accepted_data.batch,
                                mode="markers",
                                marker={"color": "red", "size": 9},
                                name="rejected",
                            ),
                        ]
                    )

            return {
                "data": traces,
                "layout": dict(
                    xaxis={"title": "Batch", "tickformat": ",d"},
                    yaxis={"title": "Objective function value"},
                ),
            }



from functools import partial

import pandas
import numpy


def calculate_statistics(data, keys):
    if keys is None or len(keys) == 0:
        raise AssertionError("List of summary keys is required")
    # Make rows out of the keys for statistics.
    data = pandas.melt(
        data,
        id_vars=["batch", "date", "realization"],
        value_vars=keys,
        var_name="summary_key",
        value_name="value",
    ).dropna()
    # Aggregate the values over the realizations, using pivot_table, keeping the
    # key, batch and date as the multi-index, calculating statistics of the
    # values over the realizations.
    index = ["summary_key", "batch", "date"]
    statistics = pandas.pivot_table(
        data,
        values="value",
        index=index,
        aggfunc=[
            numpy.mean,
            partial(numpy.quantile, q=0.1),
            partial(numpy.quantile, q=0.9),
            numpy.min,
            numpy.max,
        ],
    ).droplevel(1, axis="columns")
    min_index = data.groupby(index)["value"].idxmin()
    max_index = data.groupby(index)["value"].idxmax()
    min_realization = data.loc[min_index].set_index(index)["realization"]
    max_realization = data.loc[max_index].set_index(index)["realization"]
    result = pandas.concat([statistics, min_realization, max_realization], axis=1)
    result.columns = [
        "mean",
        "P10",
        "P90",
        "min_value",
        "max_value",
        "min_realization",
        "max_realization",
    ]
    return result.sort_index().reset_index()



from everviz.plugins.delta_plot.delta_plot import DeltaPlot
from everviz.pages.deltaplot import _get_objective_delta_values

_OBJECTIVES = [
    {"batch": 0, "realization": 1, "function": "f0", "value": 100, "simulation": 1},
    {"batch": 0, "realization": 2, "function": "f0", "value": 200, "simulation": 2},
    {"batch": 0, "realization": 1, "function": "f1", "value": 300, "simulation": 1},
    {"batch": 0, "realization": 2, "function": "f1", "value": 400, "simulation": 2},
    {"batch": 2, "realization": 1, "function": "f0", "value": 500, "simulation": 1},
    {"batch": 2, "realization": 2, "function": "f0", "value": 600, "simulation": 2},
    {"batch": 2, "realization": 1, "function": "f1", "value": 800, "simulation": 1},
    {"batch": 2, "realization": 2, "function": "f1", "value": 900, "simulation": 2},
]


def test_objective_delta_plot_callback(app, dash_duo, mocker, caplog):
    mock_api = mocker.Mock()
    mock_api.objective_values = _OBJECTIVES
    mocker.patch(
        "everviz.plugins.delta_plot.delta_plot.get_data",
        return_value=_get_objective_delta_values(mock_api, 2),
    )

    plugin = DeltaPlot(app, "values", "first")
    app.layout = plugin.layout
    dash_duo.start_server(app)

    # Select an additional key>
    dash_duo.select_dcc_dropdown("#{}".format(plugin.key_dropdown_id), "f1")

    # Clear the dropdown, which should not cause an error.
    dash_duo.clear_input("#{}".format(plugin.key_dropdown_id))
    dash_duo.clear_input("#{}".format(plugin.key_dropdown_id))

    for record in caplog.records:
        assert record.levelname != "ERROR"



def dropdown_callback(selected_control, indexed_controls):
    if selected_control is None:
        dropdown_options = list(indexed_controls.keys())
    else:
        dropdown_options = [
            name
            for name, index in indexed_controls.items()
            if len(index) == len(indexed_controls[selected_control])
        ]
    return [{"label": i, "value": i} for i in dropdown_options]



import pluggy
from everviz import everest_hooks


class TestSpec:  # pylint: disable=too-few-public-methods
    """A hook specification namespace."""

    hookspec = pluggy.HookspecMarker("test")

    @hookspec(firstresult=True)
    def visualize_data(self, api):
        """
        :param :EverestAPI instance
        """


class TestManager(pluggy.PluginManager):
    """A testing plugin manager"""

    def __init__(self):
        super().__init__("test")
        self.add_hookspecs(TestSpec)


def test_everest_hooks_registered():
    pm = TestManager()
    try:
        pm.register(everest_hooks)
    except ValueError as err:
        if not str(err).startswith("Plugin already registered"):
            raise err
    assert any(
        hook.plugin_name.startswith("everviz")
        for hook in pm.hook.visualize_data.get_hookimpls()  # pylint: disable=no-member
    )



import os
from collections import namedtuple
import pandas as pd

from everviz.log import get_logger

DataSources = namedtuple("DataSource", ["objective_values", "total_objective_values"])

logger = get_logger()


def _objective_values_from_api(api):
    return pd.DataFrame(api.objective_values)


def _total_objective_values_from_api(api):
    return pd.DataFrame(api.single_objective_values)


def _objective_values(data):
    # Sort by the index columns.
    data = data.drop(columns=["simulation"])
    sorted_values = (
        data.set_index(["function", "batch", "realization"]).sort_index().reset_index()
    )
    return sorted_values


def _total_objective_values(data):
    # Sort by the index columns.
    sorted_values = data.set_index("batch").sort_index().reset_index()
    return sorted_values


def _set_up_data_sources(api):
    everest_folder = api.output_folder
    everviz_path = os.path.join(everest_folder, "everviz")

    logger.info("Generating objective values plot")
    data = _objective_values_from_api(api)

    objective_values_file = os.path.join(everviz_path, "objective_values.csv")
    values = _objective_values(data)
    values.to_csv(objective_values_file, index=False)
    logger.info(f"File created: {objective_values_file}")  # pylint: disable=W1203

    logger.info("Generating total objective values plot")
    data = _total_objective_values_from_api(api)

    total_objective_values_file = os.path.join(
        everviz_path, "total_objective_values.csv"
    )
    values = _total_objective_values(data)
    values.to_csv(total_objective_values_file, index=False)
    logger.info(f"File created: {total_objective_values_file}")  # pylint: disable=W1203

    return DataSources(
        objective_values=objective_values_file,
        total_objective_values=total_objective_values_file,
    )


def _single_objective_title(api):
    nr_function = len(api.objective_function_names)
    if nr_function > 1:
        return "Objective functions"
    return "Objective"


def page_layout(api):
    sources = _set_up_data_sources(api)
    return {
        "title": "Objectives",
        "content": [
            "## Objective function values",
            {
                "ObjectivesPlot": {
                    "csv_file": sources.objective_values,
                },
            },
            f"## {_single_objective_title(api)}",
            {
                "SingleObjectivesPlot": {
                    "csv_file": sources.total_objective_values,
                },
            },
        ],
    }



from everviz.util import identify_indexed_controls

_crossplot_doc = """
This plot will load all the data from everest export and all columns can be visualized in a crossplot
""".strip()
_crossplot_indexed_doc = """
This plot will load the indexed control values and they can be visualized in a crossplot
""".strip()


def page_layout(api):
    everest_export = api.everest_csv
    if everest_export is not None:
        plots = [
            _crossplot_doc,
            {
                "Crossplot": {
                    "data_path": everest_export,
                },
            },
        ]

        indexed_controls = identify_indexed_controls(api.control_names)
        if len(indexed_controls) > 0:
            plots.extend(
                [
                    _crossplot_indexed_doc,
                    {
                        "CrossplotIndexed": {
                            "data_path": everest_export,
                        }
                    },
                ]
            )

        return {
            "title": "Cross plots",
            "content": plots,
        }
    return ""



import os
import everviz
from everviz.config import (
    write_webviz_config,
    webviz_config,
    setup_default_everviz_config,
)
from everviz.util import DEFAULT_CONFIG


def test_write_webviz_config(tmpdir):
    file_name = DEFAULT_CONFIG
    config = {}
    with tmpdir.as_cwd():
        assert not os.path.exists(file_name)
        write_webviz_config(config, file_name)
        assert os.path.exists(file_name)


def test_webviz_config(mocker, monkeypatch):
    monkeypatch.setattr(
        everviz.pages.controls,
        "page_layout",
        mocker.Mock(return_value=""),
    )
    monkeypatch.setattr(
        everviz.pages.crossplot,
        "page_layout",
        mocker.Mock(return_value=""),
    )
    monkeypatch.setattr(
        everviz.pages.configuration,
        "page_layout",
        mocker.Mock(return_value=""),
    )
    monkeypatch.setattr(
        everviz.pages.objectives,
        "page_layout",
        mocker.Mock(return_value=""),
    )
    monkeypatch.setattr(
        everviz.pages.deltaplot,
        "page_layout",
        mocker.Mock(return_value=""),
    )
    monkeypatch.setattr(
        everviz.pages.summary_values,
        "page_layout",
        mocker.Mock(return_value={"title": "Summary", "content": []}),
    )
    monkeypatch.setattr(
        everviz.pages.wells_values,
        "page_layout",
        mocker.Mock(return_value={}),
    )
    monkeypatch.setattr(
        everviz.pages.gradientplot,
        "page_layout",
        mocker.Mock(return_value=""),
    )

    expected_config = {
        "title": "Everest Optimization Report",
        "pages": [
            {
                "title": "Everest",
                "content": [],
            },
            {
                "title": "Summary",
                "content": [],
            },
        ],
    }

    config = webviz_config(mocker.Mock())

    assert expected_config == config


def test_setup_default_everviz_config(mocker, monkeypatch, tmpdir):
    monkeypatch.setattr(
        everviz.config,
        "webviz_config",
        mocker.Mock(
            return_value={
                "title": "Everest Optimization Report",
                "pages": [],
            }
        ),
    )

    api_mock = mocker.Mock()
    api_mock.output_folder = "everest_output"
    expected_result = os.path.join(api_mock.output_folder, "everviz", DEFAULT_CONFIG)

    with tmpdir.as_cwd():
        assert not os.path.exists(expected_result)
        result = setup_default_everviz_config(api_mock)
        assert os.path.exists(expected_result)
        assert expected_result == result



from uuid import uuid4
from pathlib import Path

from dash import html, dcc

import plotly.express as px

from everviz.plugins.plugin_abc import EvervizPluginABC
from everviz.data.load_csv.get_data import get_data
from everviz.util import base64encode


class BestControlsPlot(EvervizPluginABC):
    """
    The BestControlsPlot class implements a plugin for Webviz, for plotting the
    best control values vs the initial values.
    """

    def __init__(self, app, csv_file):
        super().__init__()

        self.graph_id = f"graph-{uuid4()}"

        self.csv_file = csv_file
        self.set_callbacks(app)

    def add_webvizstore(self):
        return [(get_data, [{"csv_file": self.csv_file}])]

    @property
    def layout(self):
        data = get_data(self.csv_file)

        fig = px.scatter(
            data,
            y="value",
            x="control",
            color="batch",
            template="plotly_white",
            labels={"control": "Control", "value": "Control Value"},
        )
        return html.Div(
            [
                dcc.Graph(
                    id=self.graph_id,
                    config={
                        "displaylogo": False,
                        "toImageButtonOptions": {"filename": "best_controls"},
                    },
                    figure=fig,
                )
            ]
        )

    def set_callbacks(self, app):
        @app.callback(self.plugin_data_output, [self.plugin_data_requested])
        def user_download_data(data_requested):
            if data_requested:
                return {
                    "filename": Path(self.csv_file).name,
                    "content": base64encode(get_data(self.csv_file).to_csv()),
                    "mime_type": "text/csv",
                }
            return None



from itertools import cycle
import plotly.graph_objs as go

from plotly.colors import DEFAULT_PLOTLY_COLORS


def _get_statistics_lines(  # pylint: disable=too-many-locals
    data, summary_key_list, line_list, x_filter, x_key, mode="lines", multi_axis=True
):
    """
    :param data: Dataframe
    :param summary_key_list: List of summary keys to plot
    :param line_list: List of batches or dates to plot
    :param x_filter: One of date|batch
    :param x_key: Opposite of x_filter
    """
    colors = cycle(DEFAULT_PLOTLY_COLORS)
    traces = []
    for idx, key in enumerate(summary_key_list):
        # Select all data belonging to the current key.
        key_data = data.xs(key, level="summary_key", drop_level=True)

        for line in line_list:
            # Select all rows belonging the current batch or date.
            line_data = key_data.xs(line, level=x_filter, drop_level=True).reset_index()

            # Set the name of the plotted line.
            name = key
            if len(line_list) > 1:
                name += f", {x_filter}:{line}"

            # Select new color for the trace lines
            color = next(colors)
            # Make the traces, with mean, P10 and P90, shading in between.
            traces.extend(
                [
                    go.Scattergl(
                        y=line_data["P90"],
                        x=line_data[x_key],
                        mode=mode,
                        marker={"size": 10},
                        line={"color": color},
                        name=name + "(P90)",
                        yaxis=f"y{idx + 1}" if idx > 0 and multi_axis else "y",
                        showlegend=False,
                    ),
                    go.Scattergl(
                        y=line_data["P10"],
                        x=line_data[x_key],
                        mode=mode,
                        line={"color": color},
                        marker={"size": 10},
                        fill="tonexty",
                        name=name + "(P10)",
                        yaxis=f"y{idx + 1}" if idx > 0 and multi_axis else "y",
                        showlegend=False,
                    ),
                    go.Scattergl(
                        y=line_data["mean"],
                        x=line_data[x_key],
                        mode=mode,
                        marker={"size": 10},
                        line={"color": color},
                        name=name,
                        yaxis=f"y{idx + 1}" if idx > 0 and multi_axis else "y",
                        showlegend=True,
                    ),
                    go.Scattergl(
                        y=line_data["min_value"],
                        x=line_data[x_key],
                        mode="markers",
                        marker={"size": 10, "symbol": "square"},
                        line={"color": color},
                        name=key,
                        yaxis=f"y{idx + 1}" if idx > 0 and multi_axis else "y",
                        showlegend=False,
                        hovertext=[
                            f"realization: {r}" for r in line_data["min_realization"]
                        ],
                    ),
                    go.Scattergl(
                        y=line_data["max_value"],
                        x=line_data[x_key],
                        mode="markers",
                        marker={"size": 10, "symbol": "square"},
                        line={"color": color},
                        name=key,
                        showlegend=False,
                        yaxis=f"y{idx + 1}" if idx > 0 and multi_axis else "y",
                        hovertext=[
                            f"realization: {r}" for r in line_data["max_realization"]
                        ],
                    ),
                ]
            )
    return traces


def _get_data_lines(
    data, summary_key_list, line_list, x_filter, x_key, mode="markers", multi_axis=True
):
    """
    :param data: Dataframe
    :param summary_key_list: List of summary keys to plot
    :param line_list: List of batches or dates to plot
    :param x_filter: One of date|batch
    :param x_key: Opposite of x_filter
    """
    # pylint: disable=too-many-locals
    colors = cycle(DEFAULT_PLOTLY_COLORS)
    realizations = data["realization"].unique()
    traces = []
    for idx, key in enumerate(summary_key_list):
        # Select all data belonging to the current key.
        key_data = data[[key, "realization"]]
        for line in line_list:
            # Select all rows belonging the current batch or date.
            line_data = key_data.xs(line, level=x_filter, drop_level=True).reset_index()
            # Set the name of the plotted line.
            name = key
            if len(line_list) > 1:
                name += f", {x_filter}:{line}"
            # Plot each realization separately, adapting the hovertext.
            show_legend = True
            color = next(colors)
            for real in realizations:
                line = line_data[line_data["realization"].isin([real])]
                traces.append(
                    go.Scattergl(
                        y=line[key],
                        x=line[x_key],
                        mode=mode,
                        marker={"size": 10},
                        name=name,
                        showlegend=show_legend,
                        line={"color": color},
                        hovertext=f"realization:{real}",
                        yaxis=f"y{idx + 1}" if idx > 0 and multi_axis else "y",
                    )
                )
                show_legend = False
    return traces


def get_callback_func(statistics):
    if statistics == "Statistics":
        return _get_statistics_lines
    return _get_data_lines


def get_layout(summary_key_list, xaxis_title, multi_axis=True):
    """
    :param summary_key_list: List of summary keys to plot
    :param xaxis_title: X-axis title for the plot
    :return: The graph figure layout
    """
    layout = dict(
        xaxis={"title": xaxis_title},
        hovermode="closest",
    )
    colors = cycle(DEFAULT_PLOTLY_COLORS)
    for idx, key in enumerate(summary_key_list):
        color = next(colors)
        y_axis = f"yaxis{idx + 1}" if idx > 0 and multi_axis else "yaxis"
        layout.update(
            {
                y_axis: dict(
                    title=key,
                    titlefont=dict(color=color),
                    tickfont=dict(color=color),
                )
            }
        )
        if idx > 0:
            layout[y_axis].update(
                dict(
                    anchor="x",
                    overlaying="y",
                    side="right" if idx % 2 else "left",
                )
            )
    return layout



from typing import List
import dash

from dash.development.base_component import Component
from dash.dependencies import Output, Input, State
from dash import html, dcc

import webviz_core_components as wcc


def layout(uuid: str, default_options: List[str] = None) -> Component:
    return html.Div(
        [
            wcc.Select(
                id=f"multi-selector_id_{uuid}",
                multi=True,
                size=10,
                persistence=True,
                options=default_options if default_options is not None else [],
            ),
            dcc.Dropdown(
                id=f"selected_dropdown_id_{uuid}",
                value=[],
                multi=True,
                options=[],
                persistence=True,
                style={"margin-bottom": 20},
            ),
            dcc.Store(id=f"selection-store_id_{uuid}", storage_type="session"),
        ]
    )


def controller_callback(uuid: str, app: dash.Dash) -> str:
    @app.callback(
        [
            Output(f"selected_dropdown_id_{uuid}", "options"),
            Output(f"selected_dropdown_id_{uuid}", "value"),
            Output(f"multi-selector_id_{uuid}", "options"),
            Output(f"multi-selector_id_{uuid}", "value"),
            Output(f"selection-store_id_{uuid}", "data"),
        ],
        [
            Input(f"multi-selector_id_{uuid}", "value"),
            Input(f"selected_dropdown_id_{uuid}", "value"),
        ],
        [
            State(f"selected_dropdown_id_{uuid}", "options"),
            State(f"selected_dropdown_id_{uuid}", "value"),
            State(f"multi-selector_id_{uuid}", "options"),
            State(f"selection-store_id_{uuid}", "data"),
        ],
    )
    def callback_selection_list(
        list_values, _, dropdown_options, dropdown_value, list_options, selection_store
    ):
        ctx = dash.callback_context
        triggered_id = ctx.triggered[0]["prop_id"].split(".")[0]

        if not triggered_id and not selection_store:
            selection_store = {"options": list_options, "selected": []}

        if triggered_id == f"multi-selector_id_{uuid}":
            for option in list_options:
                if option["value"] in list_values:
                    selection_store["selected"].append(option)

        if triggered_id == f"selected_dropdown_id_{uuid}":
            for option in dropdown_options:
                if option["value"] not in dropdown_value:
                    selection_store["selected"].remove(option)

        list_options = [
            option
            for option in selection_store["options"]
            if option not in selection_store["selected"]
        ]
        list_value = []  # No element should be highlited as selected
        dropdown_options = selection_store["selected"]
        dropdown_value = [elem["value"] for elem in dropdown_options]
        return [
            dropdown_options,
            dropdown_value,
            list_options,
            list_value,
            selection_store,
        ]

    return f"selected_dropdown_id_{uuid}"



from datetime import datetime

import pandas

from everviz.plugins.delta_plot.delta_plot import DeltaPlot
from everviz.pages.deltaplot import _get_summary_delta_values

_SUMMARY = {
    "realization": 3 * [1] + 3 * [2] + 3 * [1] + 3 * [2],
    "simulation": 3 * [1] + 3 * [2] + 3 * [1] + 3 * [2],
    "batch": [0] * 6 + [2] * 6,
    "date": 4
    * [
        datetime(2000, 1, 1).date().isoformat(),
        datetime(2000, 2, 1).date().isoformat(),
        datetime(2000, 3, 1).date().isoformat(),
    ],
    "key1": range(1, 13),
    "key2": range(10, 130, 10),
}

_EMPTY_SUMMARY = {"realization": [], "simulation": [], "date": [], "batch": []}


def test_summary_delta_plot_callback(app, dash_duo, mocker, caplog):
    mock_api = mocker.Mock()
    mock_api.summary_values.return_value = pandas.DataFrame(_SUMMARY)

    mocker.patch(
        "everviz.plugins.delta_plot.delta_plot.get_data",
        return_value=_get_summary_delta_values(mock_api, 2),
    )

    plugin = DeltaPlot(app, "values", "none")
    app.layout = plugin.layout
    dash_duo.start_server(app)

    # Select an additional key>
    dash_duo.select_dcc_dropdown("#{}".format(plugin.key_dropdown_id), "key1")
    dash_duo.select_dcc_dropdown("#{}".format(plugin.key_dropdown_id), "key2")

    # Clear the key dropdown, which should not cause an error.
    dash_duo.clear_input("#{}".format(plugin.key_dropdown_id))
    dash_duo.clear_input("#{}".format(plugin.key_dropdown_id))

    # Add back a key.
    dash_duo.select_dcc_dropdown("#{}".format(plugin.key_dropdown_id), "key2")

    # Add a date.
    dash_duo.select_dcc_dropdown(
        "#{}".format(plugin.date_dropdown_id), datetime(2000, 2, 1).date().isoformat()
    )

    # Clear the date dropdown, which should not cause an error.
    dash_duo.clear_input("#{}".format(plugin.date_dropdown_id))
    dash_duo.clear_input("#{}".format(plugin.date_dropdown_id))

    for record in caplog.records:
        assert record.levelname != "ERROR"


def test_summary_delta_plot_empty_callback(app, dash_duo, mocker, caplog):
    mock_api = mocker.Mock()
    mock_api.summary_values.return_value = pandas.DataFrame(_EMPTY_SUMMARY)

    mocker.patch(
        "everviz.plugins.delta_plot.delta_plot.get_data",
        return_value=_get_summary_delta_values(mock_api, 2),
    )

    plugin = DeltaPlot(app, "values", "none")
    app.layout = plugin.layout
    dash_duo.start_server(app)

    for record in caplog.records:
        assert record.levelname != "ERROR"



from uuid import uuid4
from pathlib import Path

from pandas import IndexSlice

import plotly.graph_objs as go

from dash import html, dcc
from dash.dependencies import Output, Input

from everviz.plugins.plugin_abc import EvervizPluginABC
from everviz.data.load_csv.get_data import get_data
from everviz.util import base64encode
from everviz.components import list_select


class GradientPlot(EvervizPluginABC):
    """
    The GradientPlot class implements a plugin for Webviz, for plotting the
    gradient values generated during an Everest optimization.
    """

    def __init__(self, app, csv_file):
        super().__init__()

        self.graph_id = f"graph-{uuid4()}"
        self.function_dropdown_id = f"dropdown-{uuid4()}"
        self.batch_dropdown_id = f"dropdown-{uuid4()}"
        self.control_dropdown_id = f"dropdown-{uuid4()}"
        self.normalization_radio_id = f"dropdown-{uuid4()}"
        self.abs_check_id = f"dropdown-{uuid4()}"
        self.list_select_id = str(uuid4())

        self.csv_file = csv_file
        self.set_callbacks(app)

    def add_webvizstore(self):
        return [(get_data, [{"csv_file": self.csv_file}])]

    @staticmethod
    def get_options(str_list):
        return [{"label": str_elem, "value": str_elem} for str_elem in str_list]

    @property
    def layout(self):
        data = get_data(self.csv_file)

        functions = data["function"].unique()
        batches = data["batch"].unique()
        controls = data["control"].unique()
        radio_buttons = ["None", "By Function", "All"]

        function_dropdown_options = self.get_options(functions)
        batch_dropdown_options = self.get_options(batches)
        controls_dropdown_options = self.get_options(controls)
        radio_options = self.get_options(radio_buttons)

        control_elements = [
            html.Label("Functions to plot:"),
            dcc.Dropdown(
                id=self.function_dropdown_id,
                options=function_dropdown_options,
                multi=True,
                value=[functions[0]],
            ),
            html.Label("Batches to plot:"),
            dcc.Dropdown(
                id=self.batch_dropdown_id,
                options=batch_dropdown_options,
                multi=True,
                value=[batches[0]],
            ),
            html.Label("Controls to plot:"),
            list_select.layout(self.list_select_id, controls_dropdown_options),
            html.Label("Normalization:"),
            dcc.RadioItems(
                id=self.normalization_radio_id,
                options=radio_options,
                value=radio_buttons[0],
                labelStyle={"display": "inline-block"},
            ),
            dcc.Checklist(
                id=self.abs_check_id,
                options=[{"label": "Plot absolute values:", "value": "abs"}],
                style={
                    "display": "inline-block",
                    "margin-right": 8,
                    "margin-top": 20,
                },
            ),
        ]

        return html.Div(
            [
                html.Div(
                    [
                        html.Div(
                            control_elements,
                            style={
                                "width": "29%",
                                "display": "inline-block",
                                "vertical-align": "top",
                            },
                        ),
                        html.Div(
                            [
                                dcc.Graph(
                                    id=self.graph_id,
                                    config={
                                        "displaylogo": False,
                                        "toImageButtonOptions": {
                                            "filename": "controls"
                                        },
                                    },
                                )
                            ],
                            style={"width": "69%", "display": "inline-block"},
                        ),
                    ]
                ),
            ]
        )

    def set_callbacks(self, app):
        list_select_id = list_select.controller_callback(self.list_select_id, app)

        @app.callback(self.plugin_data_output, [self.plugin_data_requested])
        def user_download_data(data_requested):
            if data_requested:
                return {
                    "filename": Path(self.csv_file).name,
                    "content": base64encode(get_data(self.csv_file).to_csv()),
                    "mime_type": "text/csv",
                }
            return None

        @app.callback(
            Output(self.graph_id, "figure"),
            [
                Input(self.function_dropdown_id, "value"),
                Input(self.batch_dropdown_id, "value"),
                Input(list_select_id, "value"),
                Input(self.normalization_radio_id, "value"),
                Input(self.abs_check_id, "value"),
            ],
        )
        def update_graph(
            function_list, batch_list, control_list, normalization, absolute_values
        ):
            if function_list is None or batch_list is None or not control_list:
                return {}

            data = (
                get_data(self.csv_file)
                .set_index(["batch", "function", "control"])
                .loc[IndexSlice[batch_list, function_list, control_list], :]
            )

            if absolute_values:
                data["value"] = data["value"].abs()
            if normalization == "All":
                data["value"] /= data["value"].abs().max()
            elif normalization == "By Function":
                for function in function_list:
                    data.loc[IndexSlice[:, function], "value"] /= (
                        data.loc[IndexSlice[:, function], "value"].abs().max()
                    )
            traces = []
            for function in function_list:
                for batch in batch_list:
                    batch_data = data.xs((batch, function))
                    traces.append(
                        go.Bar(
                            y=batch_data["value"],
                            x=batch_data.index,
                            name=f"{function}: batch:{batch}",
                            showlegend=True,
                        ),
                    )
            return {
                "data": traces,
                "layout": dict(
                    xaxis={"title": "Controls", "tickformat": ",d", "automargin": True},
                    yaxis={"title": "Gradient Value"},
                ),
            }



import pandas as pd
from webviz_config.common_cache import CACHE
from webviz_config.webviz_store import webvizstore


@CACHE.memoize(timeout=CACHE.TIMEOUT)
@webvizstore
def get_data(data_path) -> pd.DataFrame:
    return pd.DataFrame(pd.read_csv(data_path, sep=None, engine="python"))



import os
import time
import dash
import pytest

import pandas as pd
from PIL import Image
import everviz
from everviz.plugins import CrossplotIndexed


def test_crossplot_layout(dash_duo, monkeypatch, mocker, tmpdir, assert_equal_images):
    app = dash.Dash(__name__)
    mock_data = pd.DataFrame(data=[[1, 2, 3], [1, 2, 3]], columns=["a", "b", "c"])
    monkeypatch.setattr(
        everviz.plugins.crossplot.crossplot_indexed,
        "get_data",
        mocker.Mock(return_value=mock_data),
    )
    monkeypatch.setattr(
        everviz.plugins.crossplot.crossplot_indexed.CrossplotIndexed,
        "set_callbacks",
        mocker.Mock(),
    )
    plugin = CrossplotIndexed(mocker.Mock(), "data_path")
    layout = plugin.layout
    app.layout = layout
    dash_duo.start_server(app)

    _REFERENCE_IMAGES = {
        (1024, 768): "crossplot_indexed_layout_headless.png",
    }

    dash_duo.driver.set_window_size(1024, 768)
    with tmpdir.as_cwd():
        # Graph can be slow to load, so we need to ensure it has loaded.
        time.sleep(10)
        dash_duo.driver.save_screenshot("example_snapshot.png")
        snapshot = Image.open("example_snapshot.png")

    if snapshot.size not in _REFERENCE_IMAGES:
        pytest.skip("No reference image for layout size: {}".format(snapshot.size))

    reference_image = Image.open(
        os.path.join(
            everviz.__path__[0],
            "..",
            "test-data",
            "crossplot",
            _REFERENCE_IMAGES[snapshot.size],
        )
    )

    assert_equal_images(reference_image, snapshot, threshold=1.0)



from pathlib import Path
from uuid import uuid4

from dash import html, dcc
from dash.dependencies import Output, Input

from everviz.plugins.plugin_abc import EvervizPluginABC
from everviz.data.load_csv.get_data import get_data


from everviz.plugins.summary_plot.util import calculate_statistics
from everviz.plugins.utils.layout.sidebar_layout import get_sidebar_layout
from everviz.plugins.summary_plot import summary_callback
from everviz.util import base64encode

_WELL_RATE_KEYS = ["WOPR", "WWIR", "WGPR", "WGIR"]


class WellsPlot(EvervizPluginABC):
    """
    WellsPlot is made to visualize well rates and well target
    rates.
    """

    def __init__(self, app, csv_file, title="Well Rates"):
        super().__init__()
        self.title = title
        self.graph_id = f"graph-{uuid4()}"
        self.dropdown_key = f"dropdown-{uuid4()}"
        self.dropdown_batch = f"dropdown-{uuid4()}"
        self.radio_target = f"radio-{uuid4()}"
        self.radio_statistics = f"radio-{uuid4()}"
        self.label_id = f"label-{uuid4()}"

        self.csv_file = csv_file
        self.set_callbacks(app)

    def add_webvizstore(self):
        return [
            (get_data, [{"csv_file": self.csv_file}]),
        ]

    @property
    def layout(self):
        data = get_data(self.csv_file).set_index(["batch", "date", "realization"])
        well_summary_keys = [
            i for i in list(data.columns.unique()) if i.split(":")[0] in _WELL_RATE_KEYS
        ]
        data = data.reset_index()
        realizations = data["realization"].unique()
        batches = list(data["batch"].unique())
        statistics_toggle = ["Data", "Statistics"]
        target_rate_toggle = ["Off", "On"]

        side_bar_config = [
            (
                "dropdown",
                {
                    "item_id": self.dropdown_key,
                    "options": well_summary_keys,
                    "multi": True,
                },
            ),
            (
                "dropdown",
                {"item_id": self.dropdown_batch, "options": batches, "multi": True},
            ),
            (
                "label",
                {
                    "item_id": self.label_id,
                    "text": f"Statistics are calculated based on {len(realizations)} realizations",
                    "style": {},
                },
            ),
            (
                "radio",
                {"item_id": self.radio_statistics, "options": statistics_toggle},
            ),
            (
                "radio",
                {
                    "title": "Target rate",
                    "item_id": self.radio_target,
                    "options": target_rate_toggle,
                },
            ),
        ]

        return html.Div(
            [
                html.H1(children=self.title, style={"textAlign": "center"}),
                html.Div(
                    [
                        html.Div(
                            [get_sidebar_layout(side_bar_config)],
                            style={
                                "width": "29%",
                                "display": "inline-block",
                                "vertical-align": "top",
                            },
                        ),
                        html.Div(
                            [
                                dcc.Graph(
                                    id=self.graph_id,
                                    config={
                                        "displaylogo": False,
                                        "toImageButtonOptions": {"filename": "wells"},
                                    },
                                )
                            ],
                            style={"width": "69%", "display": "inline-block"},
                        ),
                    ]
                ),
            ]
        )

    def set_callbacks(self, app):
        @app.callback(self.plugin_data_output, [self.plugin_data_requested])
        def user_download_data(data_requested):
            if data_requested:
                return {
                    "filename": Path(self.csv_file).name,
                    "content": base64encode(get_data(self.csv_file).to_csv()),
                    "mime_type": "text/csv",
                }
            return None

        @app.callback(
            Output(self.label_id, "style"),
            [Input(self.radio_statistics, "value")],
        )
        def radio_group_click(option):
            if option == "Statistics":
                return {}
            return {"display": "none"}

        @app.callback(
            Output(self.graph_id, "figure"),
            [
                Input(self.dropdown_key, "value"),
                Input(self.dropdown_batch, "value"),
                Input(self.radio_statistics, "value"),
                Input(self.radio_target, "value"),
            ],
        )
        def update_graph(well_keys, batch_list, statistics, target_rate):
            if well_keys is None or batch_list is None:
                return {}
            if well_keys is not None and len(well_keys) == 0:
                return {}
            if batch_list is not None and len(batch_list) == 0:
                return {}

            df = get_data(self.csv_file)

            if statistics == "Statistics":
                df = calculate_statistics(df, well_keys).set_index(
                    ["summary_key", "batch", "date"]
                )
            else:
                df = df.set_index(["batch", "date"])

            callback = summary_callback.get_callback_func(statistics)

            traces = callback(
                df, well_keys, batch_list, "batch", "date", multi_axis=False
            )
            if target_rate == "On":
                target_keys = _get_target_keys(df, well_keys)
                traces.extend(
                    callback(df, target_keys, batch_list, "batch", "date", mode="lines")
                )

            return {
                "data": traces,
                "layout": dict(
                    xaxis={"title": "Date"},
                    yaxis={"title": "Well rates"},
                    hovermode="closest",
                ),
            }


def _get_target_keys(df, summary_keys):
    target_keys = []
    for summary_key in summary_keys:
        well_key, well_name = summary_key.split(":")
        target_key = "{}T:{}".format(well_key, well_name)
        if target_key in df:
            target_keys.append(target_key)
    return target_keys
